/**
 * folding_examples.c  This library contains definitions and implementations 
 *                     of some example folding operators in C.
 */




/* Types. */
#include <folding_examples.h>
/* Standard C library headers. */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>




/** Do debug printouts if DEBUG is defined, othervise ignore debug printouts. */
#ifdef DEBUG
#define dfprintf(...) fprintf(stderr, __VA_ARGS__)
#else
#define dfprintf(...)
#endif




/** Macro for square. */
#define square(x) ((x)*(x))




/** Macros for minimum and maximum calculation. */
#define min(x, y) ((x)<(y) ? (x) : (y))
#define max(x, y) ((x)>(y) ? (x) : (y))




/**
 * The identity folding operator (simple copy) in any dimension.
 */
int identity(const int nbinsin, const double* inputpdf, 
	     const void* dummy, 
	     const int nbinsout, double* foldedpdf)
{
    if ( nbinsout!=nbinsin )
    {
	dfprintf("[identity] Number of output bins (%d) != number of input bins (%d).\n", nbinsout, nbinsin);
	return -1;
    }
    memcpy(foldedpdf, inputpdf, nbinsin*sizeof(double));
    return 0;
}




/**
 * The folding operator of convolution by Gaussian centered to zero in 
 * 1 dimension.
 */
int convolveByGauss1D(const int nbinsin, const double* inputpdf, 
		      const void* parameters, 
		      const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double sigma=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    const double normalization=1.0/sqrt(2.0*M_PI*sigma*sigma);
    int i=0, j=0;
    double x=0.0, y=0.0;
    /* This is a numerical (approximate) continuum sampling in y. */
    for ( j=0 ; j<nbinsout ; ++j )
    {
	y=(j-0.5*(nbinsout-1))*Dy;
	foldedpdf[j]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( i=0 ; i<nbinsin ; ++i )
	{
	    x=(i-0.5*(nbinsin-1))*Dx;
	    foldedpdf[j]+=normalization*exp(-0.5*square((y-x)/sigma))*inputpdf[i]*Dx;
	}
    }
    return 0;
}




/**
 * The folding operator of convolution by Gaussian centered to zero in 
 * 1 dimension + binning and truncation.
 */
int convolveByGaussAndBinning1D(const int nbinsin, const double* inputpdf, 
				const void* parameters, 
				const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double sigma=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a discrete (exact) sampling in y. */
    for ( j=0 ; j<nbinsout ; ++j )
    {
	y=(j-0.5*(nbinsout-1))*Dy;
	ybinmin=y-0.5*Dy;
	ybinmax=y+0.5*Dy;
	foldedpdf[j]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( i=0 ; i<nbinsin ; ++i )
	{
	    x=(i-0.5*(nbinsin-1))*Dx;
	    /* This would be the valid code if underflow/overflow is counted 
	       into the histogram. It does not satisfy necessary regularity 
	       conditions, so then linear iterative unfolding wouldn't work. */
	    /*foldedpdf[j]+=0.5*((j==0 ? 1.0 : erf((x-ybinmin)/(M_SQRT2*sigma)))-(j+1==nbinsout ? -1.0 : erf((x-ybinmax)/(M_SQRT2*sigma))))/Dy*inputpdf[i]*Dx;*/
	    /* This is the valid code if underflow/overflow is discarded. */
	    foldedpdf[j]+=0.5*(erf((x-ybinmin)/(M_SQRT2*sigma))-erf((x-ybinmax)/(M_SQRT2*sigma)))/Dy*inputpdf[i]*Dx;
	}
    }
    return 0;
}




/**
 * The transpose folding operator of convolution by Gaussian centered 
 * to zero in 1 dimension + binning and truncation.
 */
int transposeConvolveByGaussAndBinning1D(const int nbinsin, const double* inputpdf, 
					 const void* parameters, 
					 const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double sigma=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a numerical (approximate) continuum sampling in x. */
    for ( i=0 ; i<nbinsout ; ++i )
    {
	x=(i-0.5*(nbinsout-1))*Dx;
	foldedpdf[i]=0.0;
	/* This is a discrete (exact) integral in y. */
	for ( j=0 ; j<nbinsin ; ++j )
	{
	    y=(j-0.5*(nbinsin-1))*Dy;
	    ybinmin=y-0.5*Dy;
	    ybinmax=y+0.5*Dy;
	    /* This would be the valid code if underflow/overflow is counted 
	       into the histogram. It does not satisfy necessary regularity 
	       conditions, so then linear iterative unfolding wouldn't work. */
	    /*foldedpdf[i]+=0.5*((j==0 ? 1.0 : erf((x-ybinmin)/(M_SQRT2*sigma)))-(j+1==nbinsin ? -1.0 : erf((x-ybinmax)/(M_SQRT2*sigma))))/Dy*inputpdf[j]*Dy;*/
	    /* This is the valid code if underflow/overflow is discarded. */
	    foldedpdf[i]+=0.5*(erf((x-ybinmin)/(M_SQRT2*sigma))-erf((x-ybinmax)/(M_SQRT2*sigma)))/Dy*inputpdf[j]*Dy;
	}
    }
    return 0;
}




/**
 * The folding operator of convolution by uniform distribution centered 
 * to zero in 1 dimension.
 */
int convolveByUniform1D(const int nbinsin, const double* inputpdf, 
			const void* parameters, 
			const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double width=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    const double normalization=1.0/width;
    int i=0, j=0;
    double x=0.0, y=0.0;
    /* This is a numerical (approximate) continuum sampling in y. */
    for ( j=0 ; j<nbinsout ; ++j )
    {
	y=(j-0.5*(nbinsout-1))*Dy;
	foldedpdf[j]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( i=0 ; i<nbinsin ; ++i )
	{
	    x=(i-0.5*(nbinsin-1))*Dx;
	    foldedpdf[j]+=(fabs(y-x)<0.5*width ? normalization : 0.0)*inputpdf[i]*Dx;
	}
    }
    return 0;
}




/**
 * The folding operator of convolution by uniform distribution centered 
 * to zero in 1 dimension + histogram binning and truncation.
 */
int convolveByUniformAndBinning1D(const int nbinsin, const double* inputpdf, 
			const void* parameters, 
			const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double width=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    const double normalization=1.0/width;
    const double halfwidth=0.5*width;
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a discrete (exact) sampling in y. */
    for ( j=0 ; j<nbinsout ; ++j )
    {
	y=(j-0.5*(nbinsout-1))*Dy;
	ybinmin=y-0.5*Dy;
	ybinmax=y+0.5*Dy;
	foldedpdf[j]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( i=0 ; i<nbinsin ; ++i )
	{
	    x=(i-0.5*(nbinsin-1))*Dx;
	    foldedpdf[j]+=normalization*(max(min(halfwidth, ybinmax-x), -halfwidth)-min(max(-halfwidth, ybinmin-x), halfwidth))/Dy*inputpdf[i]*Dx;
	}
    }
    return 0;
}




/**
 * The transpose folding operator of convolution by uniform distribution 
 * centered to zero in 1 dimension + histogram binning and truncation.
 */
int transposeConvolveByUniformAndBinning1D(const int nbinsin, const double* inputpdf, 
			const void* parameters, 
			const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double width=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    const double normalization=1.0/width;
    const double halfwidth=0.5*width;
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a numerical (approximate) continuum sampling in x. */
    for ( i=0 ; i<nbinsout ; ++i )
    {
	x=(i-0.5*(nbinsout-1))*Dx;
	foldedpdf[i]=0.0;
	/* This is a discrete (exact) integral in y. */
	for ( j=0 ; j<nbinsin ; ++j )
	{
	    y=(j-0.5*(nbinsin-1))*Dy;
	    ybinmin=y-0.5*Dy;
	    ybinmax=y+0.5*Dy;
	    foldedpdf[i]+=normalization*(max(min(halfwidth, ybinmax-x), -halfwidth)-min(max(-halfwidth, ybinmin-x), halfwidth))/Dy*inputpdf[j]*Dy;
	}
    }
    return 0;
}




/**
 * The folding operator of convolution by triangle distribution centered 
 * to zero in 1 dimension.
 */
int convolveByTriangle1D(const int nbinsin, const double* inputpdf, 
			 const void* parameters, 
			 const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double halfwidth=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    const double normalization=1.0/halfwidth;
    int i=0, j=0;
    double x=0.0, y=0.0;
    /* This is a numerical (approximate) continuum sampling in y. */
    for ( j=0 ; j<nbinsout ; ++j )
    {
	y=(j-0.5*(nbinsout-1))*Dy;
	foldedpdf[j]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( i=0 ; i<nbinsin ; ++i )
	{
	    x=(i-0.5*(nbinsin-1))*Dx;
	    foldedpdf[j]+=(fabs(y-x)<halfwidth ? normalization*fabs(1.0-fabs(y-x)/halfwidth) : 0.0)*inputpdf[i]*Dx;
	}
    }
    return 0;
}




/** Helper macro for convolving by triangle function. */
#define primitiveFuncTriangle(y, halfwidth, sign) ((y)+(sign)*0.5*(y)*(y)/(halfwidth))




/**
 * The folding operator of convolution by triangle distribution centered 
 * to zero in 1 dimension + histogram binning and truncation.
 */
int convolveByTriangleAndBinning1D(const int nbinsin, const double* inputpdf, 
			 const void* parameters, 
			 const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double halfwidth=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    const double normalization=1.0/halfwidth;
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a discrete (exact) sampling in y. */
    for ( j=0 ; j<nbinsout ; ++j )
    {
	y=(j-0.5*(nbinsout-1))*Dy;
	ybinmin=y-0.5*Dy;
	ybinmax=y+0.5*Dy;
	foldedpdf[j]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( i=0 ; i<nbinsin ; ++i )
	{
	    x=(i-0.5*(nbinsin-1))*Dx;
	    foldedpdf[j]+=normalization*(primitiveFuncTriangle(min(max(min(halfwidth, ybinmax-x), -halfwidth), 0), halfwidth, 1)-primitiveFuncTriangle(min(min(max(-halfwidth, ybinmin-x), halfwidth), 0), halfwidth, 1)+primitiveFuncTriangle(max(max(min(halfwidth, ybinmax-x), -halfwidth), 0), halfwidth, -1)-primitiveFuncTriangle(max(min(max(-halfwidth, ybinmin-x), halfwidth), 0), halfwidth, -1))/Dy*inputpdf[i]*Dx;
	}
    }
    return 0;
}




/**
 * The transpose folding operator of convolution by triangle distribution 
 * centered to zero in 1 dimension + histogram binning and truncation.
 */
int transposeConvolveByTriangleAndBinning1D(const int nbinsin, const double* inputpdf, 
			 const void* parameters, 
			 const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double halfwidth=params[0];
    const double Dx=params[1];
    const double Dy=params[2];
    const double normalization=1.0/halfwidth;
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a numerical (approximate) continuum sampling in x. */
    for ( i=0 ; i<nbinsout ; ++i )
    {
	x=(i-0.5*(nbinsout-1))*Dx;
	foldedpdf[i]=0.0;
	/* This is a discrete (exact) integral in y. */
	for ( j=0 ; j<nbinsin ; ++j )
	{
	    y=(j-0.5*(nbinsin-1))*Dy;
	    ybinmin=y-0.5*Dy;
	    ybinmax=y+0.5*Dy;
	    foldedpdf[i]+=normalization*(primitiveFuncTriangle(min(max(min(halfwidth, ybinmax-x), -halfwidth), 0), halfwidth, 1)-primitiveFuncTriangle(min(min(max(-halfwidth, ybinmin-x), halfwidth), 0), halfwidth, 1)+primitiveFuncTriangle(max(max(min(halfwidth, ybinmax-x), -halfwidth), 0), halfwidth, -1)-primitiveFuncTriangle(max(min(max(-halfwidth, ybinmin-x), halfwidth), 0), halfwidth, -1))/Dy*inputpdf[j]*Dy;
	}
    }
    return 0;
}




/**
 * The folding operator by calorimeter response 
 *  + histogram binning and truncation.
 */
int foldByCalorimeterAndBinning(const int nbinsin, const double* inputpdf, 
				const void* parameters, 
				const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double a=params[0];
    const double b=params[1];
    const double c=params[2];
    const double Dx=params[3];
    const double Dy=params[4];
    double mean=0.0;
    double sigma=0.0;
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a discrete (exact) sampling in y. */
    for ( j=0 ; j<nbinsout ; ++j )
    {
	y=(j-0.5*(nbinsout-1))*Dy;
	ybinmin=y-0.5*Dy;
	ybinmax=y+0.5*Dy;
	foldedpdf[j]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( i=0 ; i<nbinsin ; ++i )
	{
	    x=(i-0.5*(nbinsin-1))*Dx;
	    mean=x;
	    sigma=(x>0.0 ? a+sqrt(b*fabs(x))+c*fabs(x) : a);
	    foldedpdf[j]+=0.5*(erf((mean-ybinmin)/(M_SQRT2*sigma))-erf((mean-ybinmax)/(M_SQRT2*sigma)))/Dy*inputpdf[i]*Dx;
	}
    }
    return 0;
}




/**
 * Folding by the transposed response function of a calorimeter 
 *  + histogram binning and truncation.
 * Parameters and return value are the same as of 
 * foldByCalorimeterAndBinning.
 */
int transposeFoldByCalorimeterAndBinning(const int nbinsin, const double* inputpdf, 
					 const void* parameters, 
					 const int nbinsout, double* foldedpdf)
{
    const double* params=(const double*)parameters;
    const double a=params[0];
    const double b=params[1];
    const double c=params[2];
    const double Dx=params[3];
    const double Dy=params[4];
    double mean=0.0;
    double sigma=0.0;
    int i=0, j=0;
    double x=0.0, y=0.0, ybinmin=0.0, ybinmax=0.0;
    /* This is a discrete (exact) sampling in y. */
    for ( i=0 ; i<nbinsout ; ++i )
    {
	x=(i-0.5*(nbinsout-1))*Dx;
	mean=x;
	sigma=(x>0.0 ? a+sqrt(b*fabs(x))+c*fabs(x) : a);
	foldedpdf[i]=0.0;
	/* This is a numerical (approximate) continuum integral in x. */
	for ( j=0 ; j<nbinsin ; ++j )
	{
	    y=(j-0.5*(nbinsin-1))*Dy;
	    ybinmin=y-0.5*Dy;
	    ybinmax=y+0.5*Dy;
	    foldedpdf[i]+=0.5*(erf((mean-ybinmin)/(M_SQRT2*sigma))-erf((mean-ybinmax)/(M_SQRT2*sigma)))/Dy*inputpdf[j]*Dy;
	}
    }
    return 0;
}




/**
 * Calculates response matrix of a folding operator for use in 
 * transposeFoldNaive.
 */
double** calculateResponseMatrix(const int nbinsin, const int nbinsout, 
				 const folding foldingop, const void* parameters)
{
    int i, j;
    double** respmatrix=NULL;
    double* sampler=NULL;
    double* samplerresult=NULL;
    respmatrix=(double**)calloc(nbinsout, sizeof(double*));
    for ( i=0 ; i<nbinsout ; ++i ) respmatrix[i]=(double*)calloc(nbinsin, sizeof(double));
    sampler=(double*)calloc(nbinsin, sizeof(double));
    samplerresult=(double*)calloc(nbinsout, sizeof(double));
    memset(sampler, '\0', nbinsin*sizeof(double));
    for ( j=0 ; j<nbinsin ; ++j )
    {
	if ( j>0 ) sampler[j-1]=0.0;
	sampler[j]=1.0;
	if ( foldingop(nbinsin, sampler, parameters, nbinsout, samplerresult) )
	{
	    dfprintf("[sampleResponseMatrix] Error on folding column %d.\n", j);
	    for ( i=0 ; i<nbinsout ; ++i ) free(respmatrix[i]);
	    free(respmatrix); respmatrix=NULL;
	    break;
	}
	for ( i=0 ; i<nbinsout ; ++i ) respmatrix[i][j]=samplerresult[i];
    }
    free(samplerresult); samplerresult=NULL;
    free(sampler); sampler=NULL;

    return respmatrix;
}




/**
 * The transpose folding operator of a given folding operator 
 * (with sampled response matrix).
 */
int transposeFoldNaive(const int nbinsin, const double* inputpdf, 
		       const void* respmatrix, 
		       const int nbinsout, double* foldedpdf)
{
    const double** responsematrix=(const double**)respmatrix;
    int i, j;
    for ( i=0 ; i<nbinsout ; ++i )
    {
	foldedpdf[i]=0.0;
	for ( j=0 ; j<nbinsin ; ++j )
	{
	    foldedpdf[i]+=responsematrix[j][i]*inputpdf[j];
	}
    }
    return 0;
}




/**
 * Deallocate calculated response matrix of a folding operator.
 */
void deallocateResponseMatrix(double** respmatrix, const int nbinsout)
{
    int i;
    for ( i=0 ; i<nbinsout ; ++i ) free(respmatrix[i]);
    free(respmatrix);
}



