/**
 * folding_examples.h  This header file contains declaration of some 
 *                     example folding operators.
 */


#ifndef __FOLDING_EXAMPLES
#define __FOLDING_EXAMPLES


/** Types. */
#include <unfold_iterative.h>


/** Provide 'extern "C"' instead of 'extern' in case of C++ compilers. */
#ifdef __cplusplus
#define EXTERN extern "C"
#else
#define EXTERN extern
#endif


/**
 * The identity folding operator (simple copy) in any dimension.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  dummy	Pointer to a dummy parameter (e.g. set it to NULL).
 * @param  nbinsout     Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (if nbinsout!=nbinsin).
 */
EXTERN int identity(const int nbinsin, const double* inputpdf, 
		    const void* dummy, 
		    const int nbinsout, double* foldedpdf);


/**
 * The folding operator of convolution by Gaussian centered to zero in 
 * 1 dimension.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the standard deviation, 
 *         		the second element is the input bin width, 
 *         		the third element is the output bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int convolveByGauss1D(const int nbinsin, 
			     const double* inputpdf, 
			     const void* parameters, 
			     const int nbinsout, 
			     double* foldedpdf);


/**
 * The folding operator of convolution by Gaussian centered to zero in 
 * 1 dimension + histogram binning and truncation.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the standard deviation, 
 *         		the second element is the input bin width, 
 *         		the third element is the output bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int convolveByGaussAndBinning1D(const int nbinsin, 
				       const double* inputpdf, 
				       const void* parameters, 
				       const int nbinsout, 
				       double* foldedpdf);


/**
 * The transpose folding operator of convolution by Gaussian centered 
 * to zero in 1 dimension + histogram binning and truncation.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the standard deviation, 
 *         		the second element is the output bin width, 
 *         		the third element is the input bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int transposeConvolveByGaussAndBinning1D(const int nbinsin, 
						const double* inputpdf, 
						const void* parameters, 
						const int nbinsout, 
						double* foldedpdf);


/**
 * The folding operator of convolution by uniform distribution centered 
 * to zero in 1 dimension.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the width, 
 *         		the second element is the input bin width, 
 *         		the third element is the output bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int convolveByUniform1D(const int nbinsin, const double* inputpdf, 
			       const void* parameters, 
			       const int nbinsout, double* foldedpdf);


/**
 * The folding operator of convolution by uniform distribution centered 
 * to zero in 1 dimension + histogram binning and truncation.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the width, 
 *         		the second element is the input bin width, 
 *         		the third element is the output bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int convolveByUniformAndBinning1D(const int nbinsin, const double* inputpdf, 
			       const void* parameters, 
			       const int nbinsout, double* foldedpdf);


/**
 * The transpose folding operator of convolution by uniform distribution 
 * centered to zero in 1 dimension + histogram binning and truncation.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the width, 
 *         		the second element is the output bin width, 
 *         		the third element is the input bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int transposeConvolveByUniformAndBinning1D(const int nbinsin, const double* inputpdf, 
			       const void* parameters, 
			       const int nbinsout, double* foldedpdf);


/**
 * The folding operator of convolution by triangle distribution centered 
 * to zero in 1 dimension.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the half width, 
 *         		the second element is the input bin width, 
 *         		the third element is the output bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int convolveByTriangle1D(const int nbinsin, const double* inputpdf, 
				const void* parameters, 
				const int nbinsout, double* foldedpdf);


/**
 * The folding operator of convolution by triangle distribution centered 
 * to zero in 1 dimension + histogram binning and truncation.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the half width, 
 *         		the second element is the input bin width, 
 *         		the third element is the output bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int convolveByTriangleAndBinning1D(const int nbinsin, const double* inputpdf, 
				const void* parameters, 
				const int nbinsout, double* foldedpdf);


/**
 * The transpose folding operator of convolution by triangle distribution 
 * centered to zero in 1 dimension + histogram binning and truncation.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of first element of a double[3], the first 
 *         		element of which is the half width, 
 *         		the second element is the output bin width, 
 *         		the third element is the input bin width.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int transposeConvolveByTriangleAndBinning1D(const int nbinsin, const double* inputpdf, 
				const void* parameters, 
				const int nbinsout, double* foldedpdf);


/**
 * The folding operator by calorimeter response. Gaussian response is assumed 
 * to a given energy E, with standard deviation 
 * sigma(E)=a+sqrt(b*E)+c*E. 
 * Parameter 'a' represents energy independent noise (e.g. pedestal width), 
 * 'b' is the coefficient of stochastic term, 'c' is the coefficient of 
 * residual term. Histogram binning effect is also included.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbins 
 *         		space point locations.
 * @param  parameters	Address of the first element of a double[5] array, 
 *         		which is the a, b and c parameter as above, and 
 *         		the last two elements are the bin widths of input 
 *         		and output histograms.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values at the same locations as of inputpdf.
 * @return errorcode	Nonzero on error (always returns zero).
 */
EXTERN int foldByCalorimeterAndBinning(const int nbinsin, const double* inputpdf, 
				       const void* parameters, 
				       const int nbinsout, double* foldedpdf);


/**
 * The transpose folding operator bt response function of a calorimeter. 
 * Histogram binning effect is also included. 
 * Parameters and return value are the same as of 
 * foldByCalorimeterAndBinning. 
 */
EXTERN int transposeFoldByCalorimeterAndBinning(const int nbinsin, const double* inputpdf, 
						const void* parameters, 
						const int nbinsout, double* foldedpdf);


/**
 * Calculates response matrix of a folding operator for use in 
 * transposeFoldNaive.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  nbinsout	Total number of bins in the folded histogram.
 * @param  foldingop	The folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @return respmatrix	The sampled response matrix of the folding operator 
 *         		(NULL pointer on error).
 */
EXTERN double** calculateResponseMatrix(const int nbinsin, const int nbinsout, 
					const folding foldingop, 
					const void* parameters);


/**
 * The transpose folding operator of a given folding operator 
 * (with sampled response matrix). Note that the transpose folding does not 
 * necessarily preserve L^1 norm (however, it does in case of convolution).
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at 
 *         		nbinsin space point locations.
 * @param  respmatrix	A double**, which is the sampled response matrix 
 *         		of the folding operator to be transposed 
 *         		(rows x cols = nbinsin x nbinsout). Can be 
 *         		obtained as a result from calculateResponseMatrix.
 * @param  nbinsout	Total number of bins in the folded histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values sampled at nbinsout locations.
 * @return errorcode	Nonzero on error.
 */
EXTERN int transposeFoldNaive(const int nbinsin, const double* inputpdf, 
			      const void* respmatrix, 
			      const int nbinsout, double* foldedpdf);


/**
 * Deallocate calculated response matrix of a folding operator.
 * 
 * @param  respmatrix	The sampled response matrix of the folding operator.
 * @param  nbinsout	Total number of bins in the folded histogram.
 */
EXTERN void deallocateResponseMatrix(double** respmatrix, const int nbinsout);


#endif /* __FOLDING_EXAMPLES */
