/**
 * unfold_bayesian.c   This is a demo library for unfolding using 
 *                     Bayesian iterative method of 
 *                     Kondor-Multhei-Schorr-d'Agostini. See e.g.:
 *                     H.N.Multhei-B.Schorr: "On an Iteravtive Method for a 
 *                     Class of Integral Equations of the First Kind" 
 *                     Math.Meth.Appl.Sci. 9 (1987) 137.
 */


#ifndef __UNFOLD_BAYESIAN
#define __UNFOLD_BAYESIAN


/* Types. */
#include <unfold_iterative.h>


/**
 * Procedurefor unfolding using Bayesian method in any dimensional 
 * probability space. The parameters and return value of this function 
 * is the following.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at 
 *         		nbins space point locations.
 * @param  inputcov	Covariance matrix of the values of inputpdf for 
 *         		for exact error propagation. Set it to NULL if 
 *         		you don't want exact error propagation. Exact error 
 *         		propagation can be expensive.
 * @param  inputerr	Errors of the values of inputpdf for approximate 
 *         		error propagation. Set it to NULL if you don't 
 *         		want aproximate error propagation with neglection 
 *         		of cross-covariances at each step (this can be 
 *         		cheaper than exact error propagation).
 * @param  foldingOp	The folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  foldingOpS	The smoother folding operator. It should be a 
 *         		convolution with a pdf of the form P(eta)*eta, 
 *         		where eta is a pdf with positive and integrable 
 *         		Fourier transform, P is the parity-reflection 
 *         		operator, and '*' is convolution. Such a good 
 *         		smoother folding operator is e.g. the convolution by 
 *         		a Gaussian pdf centered to zero. Set it to NULL 
 *         		if you do not want to apply smoother folding for 
 *         		improved convergence.
 * @param  parametersS	Additional parameters passed to the smoother 
 *         		folding operator. Set it to NULL if no smoother 
 *         		folding operator is applied.
 * @param  maxiteration	Maximal number of iterations.
 * @param  errtolerance	Maximum allowed error content (this stops iteration).
 *         		Set it to non-positive to switch off this facility.
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  unfoldedpdf	Output probability density function values at the 
 *         		same locations as inputpdf.
 * @param  unfoldedcov	Covariance matrix of the values of unfoldedpdf. Set 
 *         		it to NULL if you don't want exact error propagation.
 * @param  unfoldederr	Errors of the values of unfoldedpdf. Set it to NULL 
 *         		if you don't want aproximate error propagation 
 *         		(with neglection of cross-covariances).
 * @param  niterations	Buffer to return performed number of iterations. 
 *         		Set it to NULL if not interested in.
 * @param  errorL1	Buffer to return L1 norm estimate of error content. 
 *         		Set it to NULL if not interested in.
 * @param  dAlembertL1  Buffer to return L^1 d'Alembert convergence factors 
 *         		into a buffer pre-allocated to be able to contain 
 *         		maxiteration number of values. Set it to NULL if 
 *         		you are not interested in the L^1 d'Alembert factors.
 * @param  dAlembertL2  Buffer to return L^2 d'Alembert convergence factors 
 *         		into a buffer pre-allocated to be able to contain 
 *         		maxiteration number of values. Set it to NULL if 
 *         		you are not interested in the L^2 d'Alembert factors.
 * @param  dAlembertLi  Buffer to return L^infinity d'Alembert convergence 
 *         		factors into a buffer pre-allocated to be able to 
 *         		contain maxiteration number of values. Set it to 
 *         		NULL if you are not interested in the L^infinity 
 *         		d'Alembert factors.
 * @return errorcode	Nonzero on error (e.g. if nbinsin!=nbinsout).
 */
EXTERN int unfold_bayesian(const int nbinsin, const double *inputpdf, 
			   const double **inputcovar, const double *inputerr, 
			   const folding foldingOp, const void *parameters, 
			   const folding foldingOpS, const void *parametersS, 
			   const int maxiteration, const double errtolerance, 
			   const int nbinsout, double *unfoldedpdf, 
			   double **unfoldedcov, double *unfoldederr, 
			   int *niterations, double *errorL1, 
			   double *dAlembertL1, double *dAlembertL2, 
			   double *dAlembertLi);


#endif /* __UNFOLD_BAYESIAN */
