/**
 * deconvolve_dft.cc  This is a demo library for unfolding using dft.
 */


/* Types. */
#include <deconvolve_dft.h>
/* STL stuff. */
#include <complex>
#include <fstream>
#include <iostream>
#include <vector>
#include <algorithm>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>
/* GSL stuff. */
#include <gsl/gsl_errno.h>
#include <gsl/gsl_fft_complex.h>
#include <gsl/gsl_sf.h>
using namespace std;

/**
 * Imaginary unit.
 */
static const complex<double> I(0.0, 1.0);

/**
 * Discrete Fourier transform of a uniformly sampled frequency 
 * spectrum. Assuming compact support, this is exact if mesh density 
 * goes to infinity. In continuum limit, it gives:
 * \f$
 *    R(t) := 
 *    \int_{-Fmax}^{Fmax} \exp(\mathrm{i} 2\pi t f) P(f) \;\mathrm{d}f
 * \f$.
 * 
 * Frequency domain is assumed to be [-Fmax,Fmax] (N samples). 
 * Time domain is assumed to be [-Tmax,Tmax] with Tmax=N/(4Fmax) 
 * (N samples). The sampling is assumed in the middle of frequency or 
 * time bins, therefore dF=2Fmax/N and dT=2Tmax/N is the frequency and 
 * time step size (sampling bin size).
 * 
 * @param Fmax		Maximum frequency value.
 * @param spectrum	The frequency spectrum (complex, N elements).
 * @param timeseries	The result time series (complex, N elements).
 * @return		Nonzero on error.
 * 
 * Here, N is the number of samples (has to be power of 2).
 */
static int FFT(const double Fmax, const vector< complex<double> >& spectrum, vector< complex<double> >& timeseries)
{
    const long int N=spectrum.size();
    int ret=0;
    double log2N=::log(N)/::log(2.0);
    if ( ::fabs(::round(log2N)-log2N)>1e-8 ) { return -2; }
    timeseries.resize(N);
    for ( long int i=0 ; i<N ; ++i )
    {
	timeseries[i]=((i&1)==0 ? 1.0 : -1.0)*spectrum[i];
    }
    ret=gsl_fft_complex_radix2_inverse((double*)(&(timeseries[0])), 1, N);
    if ( ret!=GSL_SUCCESS ) ret=-1;
    for ( long int i=0 ; i<N ; ++i )
    {
	timeseries[i]*=2.0*Fmax*((i&1)==0 ? 1.0 : -1.0)*(N>=4 ? 1.0 : (N==2 ? -1.0 : I));
    }
    return ret;
}


/**
 * Inverse discrete Fourier transform of a uniformly sampled time 
 * series. Assuming compact support, this is exact if mesh density 
 * goes to infinity. In continuum limit, it gives:
 * \f$
 *    P(f) := 
 *    \int_{-Tmax}^{Tmax} \exp(-\mathrm{i} 2\pi t f) R(t) \;\mathrm{d}t
 * \f$.
 * 
 * Time domain is assumed to be [-Tmax,Tmax] (N samples). 
 * Frequency domain is assumed to be [-Fmax,Fmax] with Fmax=N/(4Tmax). 
 * (N samples). The sampling is assumed in the middle of time or 
 * frequency bins, therefore dT=2Tmax/N and dF=2Fmax/N is the time and 
 * frequency step size (sampling bin size).
 * 
 * @param Fmax		Maximum frequency value.
 * @param timeseries	The time series (complex, N elements).
 * @param spectrum	The result frequency spectrum (complex, N elements).
 * @return		Nonzero on error.
 * 
 * Here, N is the number of samples (has to be power of 2).
 */
static int invFFT(const double Fmax, const vector< complex<double> >& timeseries, vector< complex<double> >& spectrum)
{
    const long int N=timeseries.size();
    int ret=0;
    double log2N=::log(N)/::log(2.0);
    if ( ::fabs(::round(log2N)-log2N)>1e-8 ) { return -2; }
    spectrum.resize(N);
    for ( long int i=0 ; i<N ; ++i )
    {
	spectrum[i]=((i&1)==0 ? 1.0 : -1.0)*timeseries[i];
    }
    ret=gsl_fft_complex_radix2_forward((double*)(&(spectrum[0])), 1, N);
    if ( ret!=GSL_SUCCESS ) ret=-1;
    for ( long int i=0 ; i<N ; ++i )
    {
	spectrum[i]*=1.0/(2.0*Fmax)*((i&1)==0 ? 1.0 : -1.0)*(N>=4 ? 1.0 : (N==2 ? -1.0 : -I));
    }
    return ret;
}


/* Error report. */
#ifdef DEBUG
#define dfprintf(...) fprintf(stderr, __VA_ARGS__)
#else
#define dfprintf(...) 
#endif


/**
 *  Procedurefor deconvolution using dft one dimensional probability space.
 */
int deconvolve_dft(const int nbinsin, const double *inputpdf, 
		   const folding foldingOp, const void *parameters, 
		   const int nbinsout, double *unfoldedpdf)
{
    const double EPSILON=1e-6;
    if ( nbinsout!=nbinsin )
    {
	dfprintf("[deconvolve_dft] Number of output and input bins have to be equal.\n");
	return -1;
    }
    const int nbins=nbinsin;
    double log2nbins=::log(nbins)/::log(2.0);
    if ( ::fabs(::round(log2nbins)-log2nbins)>1e-8 )
    {
	dfprintf("[deconvolve_dft] Number of bins has to be power of 2.\n");
	return -2;
    }
    double Tmax=1.0;
    double DT=2.0*Tmax/nbins;
    double Fmax=nbins/(4*Tmax);
    vector< complex<double> > Inputpdf(nbins);
    vector< complex<double> > InputpdfFFT(nbins);
    vector< complex<double> > Unfoldedpdf(nbins);
    vector< complex<double> > UnfoldedpdfFFT(nbins);
    vector< complex<double> > Convolverpdf(nbins);
    vector< complex<double> > ConvolverpdfFFT(nbins);
    vector<double> ConvolverpdfReal(nbins);
    vector<double> Pulsepdf(nbins);
    memset(&Pulsepdf[0], 0, nbins*sizeof(double));
    Pulsepdf[nbins/2-1]=0.5/DT;
    Pulsepdf[nbins/2]=0.5/DT;
    if ( foldingOp(nbins, &Pulsepdf[0], parameters, nbins, &ConvolverpdfReal[0]) )
    {
	dfprintf("[deconvolve_dft] Could not fold pulse.\n");
	return -3;
    }
    for ( int i=0 ; i<nbins ; ++i )
    {
	Inputpdf[i]=inputpdf[i];
	Convolverpdf[i]=ConvolverpdfReal[i];
    }
    if ( invFFT(Fmax, Inputpdf, InputpdfFFT) )
    {
	dfprintf("[deconvolve_dft] Could not fft input.\n");
	return -4;
    }
    if ( invFFT(Fmax, Convolverpdf, ConvolverpdfFFT) )
    {
	dfprintf("[deconvolve_dft] Could not fft convolver.\n");
	return -5;
    }
    for ( int i=0 ; i<nbins ; ++i )
    {
	UnfoldedpdfFFT[i]=(abs(ConvolverpdfFFT[i])>EPSILON ? InputpdfFFT[i]/ConvolverpdfFFT[i] : 0.0);
    }
    if ( FFT(Fmax, UnfoldedpdfFFT, Unfoldedpdf) )
    {
	dfprintf("[deconvolve_dft] Could not invfft deconvolved.\n");
	return -6;
    }
    for ( int i=0 ; i<nbins ; ++i )
    {
	unfoldedpdf[i]=real(Unfoldedpdf[i]);
    }
    return 0;
}



