/**
 * Matrix inversion and linear equation solving with Gauss elimination.
 */


/* Types. */
#include <invert.h>
/* Standard C helper stuff. */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>


/* Error report. */
#ifdef DEBUG
#define dfprintf(...) fprintf(stderr, __VA_ARGS__)
#else
#define dfprintf(...)
#endif


double* allocate_vector(const int dim)
{
    double* result=(double*)calloc(dim, sizeof(double));
    return result;
}


double** allocate_squarematrix(const int dim)
{
    int i;
    double** result=(double**)calloc(dim, sizeof(double*));
    for ( i=0 ; i<dim ; ++i ) result[i]=(double*)calloc(dim, sizeof(double));
    return result;
}


void free_vector(double *v)
{
    free(v);
}


void free_squarematrix(double **m, const int dim)
{
    int i;
    for ( i=0 ; i<dim ; ++i ) free(m[i]);
    free(m);
}


int solve(const matrix_t* min, const vector_t* vin, vector_t* vout)
{
  int s1,s2,s3,dim=min->dim,highest;
  double **buff_m=allocate_squarematrix(dim);
  double *buff_v=allocate_vector(dim),bufa;

  if((vin->dim!=dim)||(vout->dim!=dim)) 
    {
      dfprintf("[solve] Not compatible dimensions.\n");
      free_vector(buff_v);
      free_squarematrix(buff_m, dim);
      return(-1);
    }
  
  for(s1=0;s1<dim;s1++)
    for(s2=0;s2<dim;s2++)
      buff_m[s1][s2]=min->d[s1][s2];

  for(s1=0;s1<dim;s1++)
    buff_v[s1]=vin->d[s1];

  for(s1=0;s1<dim;s1++)
    {
      /* looking for the highest m[s1][s1] element. */
      
      highest=s1;
      for(s2=s1+1;s2<dim;s2++)
	if(fabs(buff_m[s2][s1])>fabs(buff_m[highest][s1])) highest=s2;
      
      /* Changing the two rows */
      
      for(s2=s1;s2<dim;s2++)
	{
	  bufa=buff_m[s1][s2];
	  buff_m[s1][s2]=buff_m[highest][s2];
	  buff_m[highest][s2]=bufa;
	}
      
      bufa=buff_v[s1];
      buff_v[s1]=buff_v[highest];
      buff_v[highest]=bufa;
        
      if(buff_m[s1][s1]==0) 
	{
	  dfprintf("[solve] Zero determinant matrix.\n");
	  free_vector(buff_v);
	  free_squarematrix(buff_m, dim);
	  return(-2);
	}
    
      for(s2=s1+1;s2<dim;s2++)
	{
	  for(s3=s1+1;s3<dim;s3++)
	    buff_m[s2][s3]-=buff_m[s2][s1]/buff_m[s1][s1]*buff_m[s1][s3];
	  
	  buff_v[s2]-=buff_m[s2][s1]/buff_m[s1][s1]*buff_v[s1];
	}
    }
  
  for(s1=dim-1;s1>=0;s1--)
    {
      vout->d[s1]=buff_v[s1];
      for(s2=s1+1;s2<dim;s2++)	
	vout->d[s1]-=buff_m[s1][s2]*vout->d[s2];
      
      vout->d[s1]/=buff_m[s1][s1];
    }

  free_vector(buff_v);
  free_squarematrix(buff_m, dim);
  return(0);
}


int invert(const matrix_t* min, matrix_t* mout)
{
  int s1,s2,dim=min->dim;
  vector_t bufv,bufw;
  bufv.dim=dim;
  bufv.d=allocate_vector(bufv.dim);
  bufw.dim=dim;
  bufw.d=allocate_vector(bufw.dim);

  if(mout->dim!=dim) 
    {
      dfprintf("Not compatible dimensions.\n");
      free_vector(bufv.d);
      free_vector(bufw.d);
      return(-1);
    }
  
  for(s1=0;s1<dim;s1++)
    {
      for(s2=0;s2<dim;s2++)
	bufv.d[s2]=0;
      bufv.d[s1]=1;

      solve(min,&bufv,&bufw);

      for(s2=0;s2<dim;s2++)
	mout->d[s2][s1]=bufw.d[s2];
    }

  free_vector(bufv.d);
  free_vector(bufw.d);
  return(0);
}

