/**
 * rnd.c  Random number generation tools. Based on drand48() and 
 *        srand48() functions.
 */


/* Standard C helper stuff. */
#include <rnd.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <stdio.h>


/* Random seed. */
static unsigned int randomseed=0;


/* This function sets the random seed to a given nonzero number. */
void init_rnd(const unsigned int seed)
{
    randomseed=seed;
    randomseed=(randomseed>0 ? randomseed : 1);
    srand48(randomseed);
}


/* This function returns random seed. */
unsigned int get_randomseed()
{
    return randomseed;
}


/* This function automatically initializes the random seed from 
   the actual date if not given explicitely. */
static void auto_init_rnd()
{
    if ( !randomseed )
    {
	time_t TIME;
	time(&TIME);
	randomseed=(unsigned int)(TIME);
	randomseed=(randomseed>0 ? randomseed : 1);
	srand48(randomseed);
    }
}


/* This is the prototype function for random generation. */
double uniform(const double a, const double b)
{
    /* Uniquely initializes random seed upon first call. */
    if ( !randomseed ) auto_init_rnd();
    /* Generate a random number with drand48(). */
    return a+(b-a)*drand48();
}


double triang(const double a, const double w)
{
    double r=uniform(0.0, 1.0);
    while ( r==0 ) r=uniform(0.0, 1.0);
    return a+w*(r<0.5 ? sqrt(2.0*r)-1.0 : 1.0-sqrt(2.0-2.0*r));
}


double gauss(const double a, const double s)
{
    return a+s*sqrt(2.0*fabs(log(1.0-uniform(0.0, 1.0))))*cos(2.0*M_PI*uniform(0.0, 1.0));
}


double cauchy(const double a, const double G)
{ 
    double r=uniform(0.0, 1.0);
    while ( r==0.0 ) r=uniform(0.0, 1.0);
    return a+G*tan(M_PI*(r-0.5));
}


double exponential(const double d)
{
    return d*log(1.0/(1.0-uniform(0.0, 1.0)));
}


double idexponential(const double d)
{
    return d*log(1.0/((1.0-uniform(0.0, 1.0))*(1.0-uniform(0.0, 1.0))));
}


/** Invert a distribution function by Newton-Raphson iteration.
 * @param  x       The point where the function shall be inverted.
 * @param  F       The distribution function to be inverted.
 * @param  Fprime  The derivative of the distribution function.
 * @param  param   The parameters of the function F.
 * @param  guess   The initial guess for F^{-1}(x).
 * @param  EPSILON Error tolerance.
 * @param  maxIter Maximum number of iterations.
 * @return         The estimated value of F^{-1}(x).
 */
static double newtonraphsoninverse(const double x, 
				   double (*F)(const double, const void*), 
				   double (*Fprime)(const double, const void*), 
				   const void *param, 
				   const double guess, 
				   const double EPSILON,
				   const int maxIter)
{
    double result=guess;
    double numerator=F(result, param)-x;
    int i;
    for ( i=0 ; i<maxIter ; ++i )
    {
	if ( fabs(numerator)<EPSILON ) break;
	result-=numerator/Fprime(result, param);
	numerator=F(result, param)-x;
    }
    if ( i>=maxIter ) { fprintf(stderr, "[newtonraphsoninverse] Could not reach EPSILON=%g with maxIter=%d (delta=%g, x=%g).\n", EPSILON, maxIter, fabs(numerator), x); }
    return result;
}


/* Square function. */
#define square(x) ((x)*(x))
/* The distribution function of the Tsallis based energy distribution. */
static double tsallisEnergyDistribution(const double E, const void *param)
{
    const double* params=(const double*)param;
    const double T=params[0];
    const double n=params[1];
    return 1.0-(1.0+(n-1.0)*E/(n*T))*pow(1.0+E/(n*T), -n+1.0);
}
/* The ddensity function of the Tsallis based energy distribution. */
static double tsallisEnergyDensity(const double E, const void *param)
{
    const double* params=(const double*)param;
    const double T=params[0];
    const double n=params[1];
    return E*(n-1.0)*(n-2.0)/square(n*T)*pow(1.0+E/(n*T), -n);
}


double idtsallis(const double T, const double n)
{
    const double param[]={T, n};
    const double p=uniform(0.0, 1.0);
    const double guess=-T*log(1.0-p);
    return newtonraphsoninverse(p, tsallisEnergyDistribution, tsallisEnergyDensity, param, guess, 1e-10, 100);
}

