/**
 * unfold_bayesian.c   This is a demo library for unfolding using 
 *                     Bayesian iterative method of 
 *                     Definitions and implementations in C.
 */


/* Types. */
#include <unfold_bayesian.h>
/* Standard C helper stuff. */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>


/** Do debug printouts if DEBUG is defined, othervise ignore debug printouts. */
#ifdef DEBUG
#define dfprintf(...) fprintf(stderr, __VA_ARGS__)
#else
#define dfprintf(...)
#endif


/** Allocate vector. */
static double* allocatevector(const int dim)
{
    double* result=(double*)calloc(dim, sizeof(double));
    return result;
}


/** Deallocate vector. */
static void freevector(double *v)
{
    free(v);
}


/** Allocate square matrix. */
static double** allocatesquarematrix(const int dim)
{
    int i;
    double** result=(double**)calloc(dim, sizeof(double*));
    for ( i=0 ; i<dim ; ++i ) result[i]=(double*)calloc(dim, sizeof(double));
    return result;
}


/** Deallocate square matrix. */
static void freesquarematrix(double **m, const int dim)
{
    int i;
    for ( i=0 ; i<dim ; ++i ) free(m[i]);
    free(m);
}


/** Set vector to zero. */
static void annulatevector(const int dim, double *vector)
{
    memset(vector, '\0', dim*sizeof(double));
}


/** Copy a vector. */
static void copyvector(const int dim, const double *sourcevector, double *destvector)
{
    memcpy(destvector, sourcevector, dim*sizeof(double));
}


/** Copy a square matrix. */
static void copysquarematrix(const int dim, const double **sourcematrix, double **destmatrix)
{
    int i=0;
    for ( i=0 ; i<dim ; ++i ) memcpy(destmatrix[i], sourcematrix[i], dim*sizeof(double));
}


/** Transpose a square matrix. */
static void transposesquarematrix(const int dim, double **matrix)
{
    int i=0, j=0;
    double bufferij=0.0;
    for ( i=0 ; i<dim ; ++i )
    {
	for ( j=i+1 ; j<dim ; ++j )
	{
	    bufferij=matrix[i][j];
	    matrix[i][j]=matrix[j][i];
	    matrix[j][i]=bufferij;
	}
    }
}


/**
 * Apply folding operator to a square matrix as vector of histograms.
 * 
 * @param  nbins	Total number of bins along one edge of the 
 *         		input/output matrix.
 * @param  inputmat	Input matrix.
 * @param  foldingOp	Folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  foldedmat	Buffer to return output matrix.
 * @return errorcode	Nonzero on error.
 */
static int applytosquarematrix(const int nbins, const double **inputmat, 
			       const folding foldingOp, const void *parameters, 
			       double **foldedmat)
{
    /* Error code. */
    int errorcode=0;
    /* Indices. */
    int i=0, j=0;
    /* Buffers. */
    double *sampler=NULL;
    double *samplerresult=NULL;

    /* Allocate memory. */
    sampler=allocatevector(nbins);
    samplerresult=allocatevector(nbins);

    /* Apply folding to each column. */
    for ( j=0 ; j<nbins ; ++j )
    {
	for ( i=0 ; i<nbins ; ++i ) sampler[i]=inputmat[i][j];
	if ( foldingOp(nbins, sampler, parameters, nbins, samplerresult) )
	{
	    dfprintf("[applytosquarematrix] Failed to apply folding to %d-th column of matrix.\n", j);
	    errorcode=-1;
	    break;
	}
	for ( i=0 ; i<nbins ; ++i ) foldedmat[i][j]=samplerresult[i];
    }

    /* Deallocate memory. */
    freevector(sampler); sampler=NULL;
    freevector(samplerresult); samplerresult=NULL;

    return errorcode;
}


/**
 * Sample folding operator as a square matrix.
 * 
 * @param  nbins	Total number of bins along one edge of the 
 *         		output matrix.
 * @param  foldingOp	Folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  matrix	Buffer to return output matrix.
 * @return errorcode	Nonzero on error.
 */
static int sampleasmatrix(const int nbins, const folding foldingOp, 
			  const void *parameters, double **matrix)
{
    /* Error code. */
    int errorcode=0;
    /* Indices. */
    int i=0, j=0;
    /* Temporary vector for sampling the folding operator. */
    double* sampler=NULL;
    double* samplerresult=NULL;

    /* Initialize folder matrix. */
    sampler=allocatevector(nbins);
    samplerresult=allocatevector(nbins);
    annulatevector(nbins, sampler);
    for ( i=0 ; i<nbins ; ++i )
    {
	if ( i>0 ) sampler[i-1]=0.0;
	sampler[i]=1.0;
	if ( foldingOp(nbins, sampler, parameters, nbins, samplerresult) )
	{
	    dfprintf("[sampleasmatrix] Error on folding %d.\n", i);
	    errorcode=-1;
	    break;
	}
	for ( j=0 ; j<nbins ; ++j ) matrix[j][i]=samplerresult[j];
    }
    freevector(samplerresult); samplerresult=NULL;
    freevector(sampler); sampler=NULL;

    return errorcode;
}


/**
 * Type for parameter to a composition of folding operators (max 4).
 */
typedef struct
{
    /* Number of compositions (max 4). */
    int ncomposition;
    /* List of folding operators. */
    folding folder[4];
    /* List of parameter pointers of the consecutive folding operators. */
    void *parameter[4];
    /* List of output bin numbers of the consecutive folding operators. */
    int nbinsout[4];
}
paramcompose_t;


/**
 * Folding operator composed of folding operators.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbinsin 
 *         		space point locations.
 * @param  parameters	Additional parameters passed to the folding operator 
 *         		(pointer to a paramcompose_t object).
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values sampled at nbinsout locations.
 * @return errorcode	Nonzero on error.
 */
static int compose(const int nbinsin, const double* inputpdf, 
		   const void* parameters, 
		   const int nbinsout, double* foldedpdf)
{
    /* Error code. */
    int errorcode=0;
    /* Parameters. */
    const paramcompose_t *paramcompose=(const paramcompose_t*)parameters;
    const int ncomposition=paramcompose->ncomposition;
    /* Indices. */
    int n=0;
    /* Number of bins. */
    int nbinsinbuff=0;
    int nbinsoutbuff=0;
    int nmaxbins=0;
    /* Buffers. */
    double* buff[2]={NULL, NULL};
    double* buffpointer[2]={NULL, NULL};
    int inputbuff=0;

    /* If empty composition, return input at output. */
    if ( ncomposition==0 )
    {
	if ( nbinsout!=nbinsin ) return -1;
	copyvector(nbinsin, inputpdf, foldedpdf);
	return 0;
    }

    /* Get maximal necessary size of internal buffers. */
    for ( n=0 ; n<ncomposition-1 ; ++n )
    {
	if ( nmaxbins<paramcompose->nbinsout[n] ) nmaxbins=paramcompose->nbinsout[n];
    }

    /* Loop over compositions. */
    for ( n=0 ; n<ncomposition ; ++n )
    {
	/* If we are at the first, take input as input buffer. */
	if ( n==0 ) { nbinsinbuff=nbinsin; buffpointer[inputbuff]=(double*)inputpdf; }
	else { nbinsinbuff=paramcompose->nbinsout[n-1]; buffpointer[inputbuff]=buff[inputbuff]; }
	/* If we are at the last, take output as output buffer. */
	if ( n+1==ncomposition ) { nbinsoutbuff=nbinsout; buffpointer[!inputbuff]=foldedpdf; }
	else { nbinsoutbuff=paramcompose->nbinsout[n]; buffpointer[!inputbuff]=buff[!inputbuff]; }
	/* Allocate buffers when necessary. */
	if ( buffpointer[inputbuff]==NULL ) { buff[inputbuff]=allocatevector(nmaxbins); buffpointer[inputbuff]=buff[inputbuff]; }
	if ( buffpointer[!inputbuff]==NULL ) { buff[!inputbuff]=allocatevector(nmaxbins); buffpointer[!inputbuff]=buff[!inputbuff]; }
	/* Fold successively. */
	if ( (paramcompose->folder[n])(nbinsinbuff, buffpointer[inputbuff], paramcompose->parameter[n], nbinsoutbuff, buffpointer[!inputbuff]) )
	{
	    dfprintf("[compose] Error on %d-th folding composition.\n", n);
	    errorcode=-1;
	    break;
	}
	/* Swap input/output buffers. */
	inputbuff=!inputbuff;
    }

    /* Free buffers. */
    if ( buff[inputbuff]!=NULL ) { freevector(buff[inputbuff]); buff[inputbuff]=NULL; }
    if ( buff[!inputbuff]!=NULL ) { freevector(buff[!inputbuff]); buff[!inputbuff]=NULL; }

    return errorcode;
}


/**
 * Procedurefor Bayesian unfolding in any dimensional 
 * probability space.
 */
int unfold_bayesian(const int nbinsin, const double *inputpdf, 
		    const double **inputcov, const double *inputerr, 
		    const folding foldingOp, const void *parameters, 
		    const folding foldingOpS, const void *parametersS, 
		    const int maxiteration, const double errtolerance, 
		    const int nbinsout, double *unfoldedpdf, 
		    double **unfoldedcov, 
		    double *unfoldederr, 
		    int *niterations, double *errorL1, 
		    double *dAlembertL1, double* dAlembertL2, 
		    double *dAlembertLi)
{
    const double EPSILON=1e-8;
    /* Error code. */
    int errorcode=0;
    /* Number of bins. */
    const int nbins=nbinsin;
    /* Indices. */
    int i=0, j=0, n=0;
    /* Initial composition parameters. */
    paramcompose_t paramcomposeinitial;
    /* Composition parameters during iteration. */
    paramcompose_t paramcompose;
    /* L^1 norm of initial pdf for error tolerance reference. */
    double L1norm=0.0;
    /* d'Alembert factor components. */
    double dAlembertl1=1.0;
    double dAlembertl1prev=1.0;
    double dAlembertl2=1.0;
    double dAlembertl2prev=1.0;
    double dAlembertli=1.0;
    double dAlembertliprev=1.0;
    /* L^1 norm of error content. */
    double errL1=0.0;
    /* Buffers for unfolded pdf. */
    double *f0=NULL;
    double *fn=NULL;
    double *rf=NULL;
    /* Buffers for propagated covariance of unfolded pdf. */
    double **c0=NULL;
    double **cn=NULL;
    double **cb=NULL;
    double **rc=NULL;
    /* Buffers for estimated variance of unfolded pdf. */
    double *v0=NULL;
    double *vn=NULL;
    double *vf=NULL;
    /* Buffer for matrix of folding. */
    double **ma=NULL;

    if ( nbinsout!=nbinsin )
    {
	dfprintf("[unfold_bayesian] nbinsout!=nbinsin.\n");
	return -1;
    }

    /* Calculate L^1 norm for our reference for error tolerance. */
//    if ( errtolerance>0.0 && ( inputcov!=NULL || inputerr!=NULL ) )
//    {
	L1norm=0.0;
	for ( i=0 ; i<nbins ; ++i ) L1norm+=fabs(inputpdf[i]);
//    }

    /* Prepare composite operators. */
    paramcomposeinitial.ncomposition=0;
    paramcompose.ncomposition=0;
    if ( foldingOp!=NULL )
    {
	paramcompose.folder[paramcompose.ncomposition]=foldingOp;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parameters;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsin;
	++paramcompose.ncomposition;
    }
    if ( foldingOpS!=NULL )
    {
	paramcomposeinitial.folder[paramcomposeinitial.ncomposition]=foldingOpS;
	paramcomposeinitial.parameter[paramcomposeinitial.ncomposition]=(void*)parametersS;
	paramcomposeinitial.nbinsout[paramcomposeinitial.ncomposition]=nbinsin;
	++paramcomposeinitial.ncomposition;
	paramcompose.folder[paramcompose.ncomposition]=foldingOpS;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parametersS;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsin;
	++paramcompose.ncomposition;
    }

    /* Prepare initial value. */
    f0=allocatevector(nbins);
    fn=allocatevector(nbins);
    if ( compose(nbins, inputpdf, &paramcomposeinitial, nbins, f0) )
    {
	dfprintf("[unfold_bayesian] Failed to prepare initial value.\n");
	if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	if ( rf!=NULL ) { freevector(rf); rf=NULL; }
	if ( c0!=NULL ) { freesquarematrix(c0, nbins); c0=NULL; }
	if ( cn!=NULL ) { freesquarematrix(cn, nbins); cn=NULL; }
	if ( cb!=NULL ) { freesquarematrix(cb, nbins); cb=NULL; }
	if ( rc!=NULL ) { freesquarematrix(rc, nbins); rc=NULL; }
	if ( v0!=NULL ) { freevector(v0); v0=NULL; }
	if ( vn!=NULL ) { freevector(vn); vn=NULL; }
	if ( vf!=NULL ) { freevector(vf); vf=NULL; }
	if ( ma!=NULL ) { freesquarematrix(ma, nbins); ma=NULL; }
	return -2;
    }
    copyvector(nbins, f0, fn);
    if ( inputcov!=NULL )
    {
	c0=allocatesquarematrix(nbins);
	cb=allocatesquarematrix(nbins);
	cn=allocatesquarematrix(nbins);
	if ( applytosquarematrix(nbins, inputcov, compose, &paramcomposeinitial, cb) )
	{
	    dfprintf("[unfold_bayesian] Failed to prepare initial covariance 1.\n");
	    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	    if ( rf!=NULL ) { freevector(rf); rf=NULL; }
	    if ( c0!=NULL ) { freesquarematrix(c0, nbins); c0=NULL; }
	    if ( cn!=NULL ) { freesquarematrix(cn, nbins); cn=NULL; }
	    if ( cb!=NULL ) { freesquarematrix(cb, nbins); cb=NULL; }
	    if ( rc!=NULL ) { freesquarematrix(rc, nbins); rc=NULL; }
	    if ( v0!=NULL ) { freevector(v0); v0=NULL; }
	    if ( vn!=NULL ) { freevector(vn); vn=NULL; }
	    if ( vf!=NULL ) { freevector(vf); vf=NULL; }
	    if ( ma!=NULL ) { freesquarematrix(ma, nbins); ma=NULL; }
	    return -3;
	}
	transposesquarematrix(nbins, cb);
	if ( applytosquarematrix(nbins, (const double **)cb, compose, &paramcomposeinitial, c0) )
	{
	    dfprintf("[unfold_bayesian] Failed to prepare initial covariance 2.\n");
	    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	    if ( rf!=NULL ) { freevector(rf); rf=NULL; }
	    if ( c0!=NULL ) { freesquarematrix(c0, nbins); c0=NULL; }
	    if ( cn!=NULL ) { freesquarematrix(cn, nbins); cn=NULL; }
	    if ( cb!=NULL ) { freesquarematrix(cb, nbins); cb=NULL; }
	    if ( rc!=NULL ) { freesquarematrix(rc, nbins); rc=NULL; }
	    if ( v0!=NULL ) { freevector(v0); v0=NULL; }
	    if ( vn!=NULL ) { freevector(vn); vn=NULL; }
	    if ( vf!=NULL ) { freevector(vf); vf=NULL; }
	    if ( ma!=NULL ) { freesquarematrix(ma, nbins); ma=NULL; }
	    return -4;
	}
	copysquarematrix(nbins, (const double **)c0, cn);
    }
    else if ( inputerr!=NULL )
    {
	v0=allocatevector(nbins);
	vn=allocatevector(nbins);
	ma=allocatesquarematrix(nbins);
	if ( sampleasmatrix(nbins, compose, &paramcomposeinitial, ma) )
	{
	    dfprintf("[unfold_bayesian] Failed to sample as matrix initially.\n");
	    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	    if ( rf!=NULL ) { freevector(rf); rf=NULL; }
	    if ( c0!=NULL ) { freesquarematrix(c0, nbins); c0=NULL; }
	    if ( cn!=NULL ) { freesquarematrix(cn, nbins); cn=NULL; }
	    if ( cb!=NULL ) { freesquarematrix(cb, nbins); cb=NULL; }
	    if ( rc!=NULL ) { freesquarematrix(rc, nbins); rc=NULL; }
	    if ( v0!=NULL ) { freevector(v0); v0=NULL; }
	    if ( vn!=NULL ) { freevector(vn); vn=NULL; }
	    if ( vf!=NULL ) { freevector(vf); vf=NULL; }
	    if ( ma!=NULL ) { freesquarematrix(ma, nbins); ma=NULL; }
	    return -5;
	}
	for ( i=0 ; i<nbins ; ++i )
	{
	    v0[i]=0.0;
	    for ( j=0 ; j<nbins ; ++j ) v0[i]+=ma[i][j]*inputerr[j]*inputerr[j]*ma[j][i];
	}
	copyvector(nbins, v0, vn);
    }

    /* Perform iteration. */
    for ( i=0 ; i<nbins ; ++i ) //Uniform input distribution.
    {
	fn[i]=1.0/(double)nbins*L1norm;
    }
    if ( maxiteration>0 )
    {
	rf=allocatevector(nbins);
	if ( inputcov!=NULL ) rc=allocatesquarematrix(nbins);
	else if ( inputerr!=NULL )
	{
	    vf=allocatevector(nbins);
	}
	if ( ma==NULL ) ma=allocatesquarematrix(nbins);
	if ( sampleasmatrix(nbins, compose, &paramcompose, ma) )
	{
	    dfprintf("[unfold_bayesian] Failed to sample as matrix.\n");
	    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	    if ( rf!=NULL ) { freevector(rf); rf=NULL; }
	    if ( c0!=NULL ) { freesquarematrix(c0, nbins); c0=NULL; }
	    if ( cn!=NULL ) { freesquarematrix(cn, nbins); cn=NULL; }
	    if ( cb!=NULL ) { freesquarematrix(cb, nbins); cb=NULL; }
	    if ( rc!=NULL ) { freesquarematrix(rc, nbins); rc=NULL; }
	    if ( v0!=NULL ) { freevector(v0); v0=NULL; }
	    if ( vn!=NULL ) { freevector(vn); vn=NULL; }
	    if ( vf!=NULL ) { freevector(vf); vf=NULL; }
	    if ( ma!=NULL ) { freesquarematrix(ma, nbins); ma=NULL; }
	    return -6;
	}
    }
    dAlembertl1=1.0;
    dAlembertl1prev=1.0;
    dAlembertl2=1.0;
    dAlembertl2prev=1.0;
    dAlembertli=1.0;
    dAlembertliprev=1.0;
    for ( n=0 ; n<maxiteration ; ++n )
    {
	/* Calculate the n-th iteration. */
	if ( compose(nbins, fn, &paramcompose, nbins, rf) )
	{
	    dfprintf("[unfold_bayesian] Failed in iteration step %d.\n", n);
	    errorcode=-7;
	    break;
	}
	if ( inputcov==NULL && inputerr!=NULL )
	{
	    for ( i=0 ; i<nbins ; ++i )
	    {
		vf[i]=0.0;
		for ( j=0 ; j<nbins ; ++j ) vf[i]+=0.0;//FIXME!
	    }
	}
	dAlembertl1=0.0;
	dAlembertl2=0.0;
	dAlembertli=0.0;
	for ( i=0 ; i<nbins ; ++i )
	{
	    double coefficienti=0.0;
	    for ( j=0 ; j<nbins ; ++j )
	    {
		coefficienti+=(fabs(rf[j])<EPSILON ? 0.0 : (ma[j][i]*f0[j])/rf[j]);
	    }
	    double incr=fabs((coefficienti-1.0)*fn[i]);//FIXME!
	    dAlembertl1+=incr;
	    dAlembertl2+=incr*incr;
	    if ( dAlembertli<incr ) dAlembertli=incr;
	    fn[i]*=coefficienti;
	    if ( inputcov==NULL && inputerr!=NULL )
	    {
		vn[i]+=0.0;//FIXME!
	    }
	}
	dAlembertl2=sqrt(fabs(dAlembertl2));
	if ( inputcov!=NULL )
	{
	    if ( applytosquarematrix(nbins, (const double **)cn, compose, &paramcompose, cb) )
	    {
		dfprintf("[unfold_bayesian] Failed in iteration step %d for covariance 1.\n", n);
		errorcode=-8;
		break;
	    }
	    transposesquarematrix(nbins, cb);
	    if ( applytosquarematrix(nbins, (const double **)cb, compose, &paramcompose, rc) )
	    {
		dfprintf("[unfold_bayesian] Failed in iteration step %d for covariance 2.\n", n);
		errorcode=-9;
		break;
	    }
	    for ( i=0 ; i<nbins ; ++i )
	    {
		cn[i][i]+=0.0;//FIXME!
		for ( j=i+1 ; j<nbins ; ++j )
		{
		    cn[i][j]+=0.0;//FIXME!
		    cn[j][i]=cn[i][j];
		}
	    }
	}
	/* Report L^1 d'Alembert convergence factor and number of iterations. */
	if ( dAlembertL1!=NULL ) dAlembertL1[n]=(n>0 ? dAlembertl1/dAlembertl1prev : 1.0);
	if ( dAlembertL2!=NULL ) dAlembertL2[n]=(n>0 ? dAlembertl2/dAlembertl2prev : 1.0);
	if ( dAlembertLi!=NULL ) dAlembertLi[n]=(n>0 ? dAlembertli/dAlembertliprev : 1.0);
	dAlembertl1prev=dAlembertl1;
	dAlembertl2prev=dAlembertl2;
	dAlembertliprev=dAlembertli;
	if ( niterations!=NULL ) *niterations=n+1;
	/* Break loop if error content becomes too large. */
	errL1=0.0;
	if ( inputcov!=NULL )
	{
	    for ( i=0 ; i<nbins ; ++i ) errL1+=sqrt(fabs(cn[i][i]));
	}
	else if ( inputerr!=NULL )
	{
	    for ( i=0 ; i<nbins ; ++i ) errL1+=sqrt(fabs(vn[i]));
	}
	if ( inputcov!=NULL || inputerr!=NULL )
	{
	    errL1/=L1norm;
	    if ( errorL1!=NULL ) *errorL1=errL1;
	    if ( errtolerance>0.0 && errL1>errtolerance ) break;
	}
    }
    if ( unfoldedpdf!=NULL ) copyvector(nbins, fn, unfoldedpdf);
    if ( inputcov!=NULL && unfoldedcov!=NULL ) copysquarematrix(nbins, (const double **)cn, unfoldedcov);
    else if ( inputerr!=NULL && unfoldederr!=NULL )
    {
	for ( i=0 ; i<nbins ; ++i ) unfoldederr[i]=sqrt(fabs(vn[i]));
    }

    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
    if ( rf!=NULL ) { freevector(rf); rf=NULL; }
    if ( c0!=NULL ) { freesquarematrix(c0, nbins); c0=NULL; }
    if ( cn!=NULL ) { freesquarematrix(cn, nbins); cn=NULL; }
    if ( cb!=NULL ) { freesquarematrix(cb, nbins); cb=NULL; }
    if ( rc!=NULL ) { freesquarematrix(rc, nbins); rc=NULL; }
    if ( v0!=NULL ) { freevector(v0); v0=NULL; }
    if ( vn!=NULL ) { freevector(vn); vn=NULL; }
    if ( vf!=NULL ) { freevector(vf); vf=NULL; }
    if ( ma!=NULL ) { freesquarematrix(ma, nbins); ma=NULL; }

    return errorcode;
}



