/**
 * unfold_inverse.c   This is a demo library for unfolding using 
 *                    matrix inversion.
 */


/* Types. */
#include <unfold_inverse.h>
/* The invert library. */
#include <invert.h>
/* Standard C helper stuff. */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>


/* Error report. */
#ifdef DEBUG
#define dfprintf(...) fprintf(stderr, __VA_ARGS__)
#else
#define dfprintf(...) 
#endif


/**
 * Procedurefor unfolding using matrix inversion in any dimensional 
 * probability space.
 */
int unfold_inverse(const int nbinsin, const double *inputpdf, 
		   const folding foldingOp, const void *parameters, 
		   const int nbinsout, double *unfoldedpdf)
{
    /* Number of bins. */
    const int nbins=nbinsin;
    /* Indices. */
    int i=0, j=0;

    /* Folder matrix. */
    matrix_t folder;
    /* Temporary vector for sampling the folding operator. */
    vector_t sampler;
    vector_t samplerresult;
    /* Vector of folding result. */
    vector_t foldingresult;
    /* Vector of uknown folding argument. */
    vector_t foldingargument;

    if ( nbinsout!=nbinsin )
    {
	dfprintf("[unfold_inverse] nbinsout!=nbinsin\n");
	return -1;
    }

    /* Allocate memory for matrix buffers. */
    folder.dim=nbins;
    folder.d=allocate_squarematrix(nbins);
    foldingresult.dim=nbins;
    foldingresult.d=(double*)inputpdf;
    foldingargument.dim=nbins;
    foldingargument.d=unfoldedpdf;

    /* Initialize folder matrix. */
    sampler.dim=nbins;
    sampler.d=allocate_vector(nbins);
    samplerresult.dim=nbins;
    samplerresult.d=allocate_vector(nbins);
    memset(sampler.d, 0, nbins*sizeof(double));
    for ( i=0 ; i<nbins ; ++i )
    {
	if ( i>0 ) sampler.d[i-1]=0.0;
	sampler.d[i]=1.0;
	if ( foldingOp(nbins, sampler.d, parameters, nbins, samplerresult.d) )
	{
	    dfprintf("[unfold_inverse] Error on folding %d.\n", i);
	    free_vector(samplerresult.d);
	    free_vector(sampler.d);
	    free_squarematrix(folder.d, nbins);
	    return -2;
	}
	for ( j=0 ; j<nbins ; ++j ) folder.d[j][i]=samplerresult.d[j];
    }
    free_vector(samplerresult.d);
    free_vector(sampler.d);

    /* Solve linear equation 
       foldermatrix foldingargument = foldingresult
       with Gauss elimination. */
    if ( solve(&folder, &foldingresult, &foldingargument) )
    {
	dfprintf("[unfold_inverse] Error on solving.\n");
	free_squarematrix(folder.d, nbins);
	return -3;
    }

    /* Deallocate buffer memories for matrices. */
    free_squarematrix(folder.d, nbins);

    return 0;
}




/**
 * Procedureto determine matrix norm of an unfolding operator in a given 
 * discretization in any dimensional probability space.
 */
int unfold_inverseoperatornorm(const int nbinsin, const folding foldingOp, 
			       const void *parameters, const int nbinsout, 
			       double *opNorm)
{
    /* Number of bins. */
    const int nbins=nbinsin;
    /* Indices. */
    int i=0, j=0;

    /* Folder matrix. */
    matrix_t folder;
    /* Temporary vector for sampling the folding operator. */
    vector_t sampler;
    vector_t samplerresult;
    /* Inverse folder matrix. */
    matrix_t unfolder;

    /* Set operatornorm to zero. */
    *opNorm=0.0;

    if ( nbinsout!=nbinsin )
    {
	dfprintf("[unfold_inverse] nbinsout!=nbinsin\n");
	return -1;
    }

    /* Allocate memory for matrix buffers. */
    folder.dim=nbins;
    folder.d=allocate_squarematrix(nbins);

    /* Initialize folder matrix. */
    sampler.dim=nbins;
    sampler.d=allocate_vector(nbins);
    samplerresult.dim=nbins;
    samplerresult.d=allocate_vector(nbins);
    memset(sampler.d, 0, nbins*sizeof(double));
    for ( i=0 ; i<nbins ; ++i )
    {
	if ( i>0 ) sampler.d[i-1]=0.0;
	sampler.d[i]=1.0;
	if ( foldingOp(nbins, sampler.d, parameters, nbins, samplerresult.d) )
	{
	    dfprintf("[unfold_inverse] Error on folding %d.\n", i);
	    free_vector(samplerresult.d);
	    free_vector(sampler.d);
	    free_squarematrix(folder.d, nbins);
	    return -2;
	}
	for ( j=0 ; j<nbins ; ++j ) folder.d[j][i]=samplerresult.d[j];
    }
    free_vector(samplerresult.d);
    free_vector(sampler.d);

    /* Invert the folder matrix and put result to unfolder matrix. */
    unfolder.dim=nbins;
    unfolder.d=allocate_squarematrix(nbins);
    if ( invert(&folder, &unfolder) )
    {
	dfprintf("[unfold_operatornorm] Error on inversion.\n");
	free_squarematrix(folder.d, nbins);
	free_squarematrix(unfolder.d, nbins);
	return -3;
    }
    free_squarematrix(folder.d, nbins);

    /* Calculate L^1-induced matrix norm of unfolder matrix. */
    double matrixnorm=0.0;
    for ( j=0 ; j<nbins ; ++j )
    {
	double l1normcolumn=0.0;
	for ( i=0 ; i<nbins ; ++i )
	{
	    double absolutevalue=fabs(unfolder.d[i][j]);
	    l1normcolumn+=absolutevalue;
	}
	if ( matrixnorm<l1normcolumn ) matrixnorm=l1normcolumn;
    }

    /* Deallocate buffer memories. */
    free_squarematrix(unfolder.d, nbins);

    /* Report result. */
    *opNorm=matrixnorm;

    return 0;
}

