////////////////////////////////////////////////////////////////////////////
// unfold_demo.cc  This is a demo for unfolding techniques.               //
////////////////////////////////////////////////////////////////////////////


#include <rnd.h>
#include <deconvolve_dft.h>
#include <unfold_inverse.h>
#include <unfold_bayesian.h>
#include <unfold_iterative.h>
#include <folding_examples.h>
#include <cstdlib>
#include <float.h>
#include <cmath>
#include <cstring>
#include <vector>
#include <iostream>
#include <fstream>
using namespace std;


// The lower border.
const double lowerborder=-16.0;
// The upper border.
const double upperborder=16.0;
// The discretization of the problem.
const int nbins=64;
// The folded pdf / initial pdf domain width ratio (odd).
const int widthratio=3;
// The folded pdf / initial pdf samplig ratio (odd).
const int samplingratio=9;
// Width parameter of the input Cauchy pdf.
const double Gamma=1.0;
// Generate measured pdf with Monte Carlo.
const bool withMC=true;
// Number of entries in the input histogram.
const int Entries=5000;
// Type of convolver function.
enum type { Gauss=0, Uniform=1, Triangle=2 };
type Type=Gauss;
// Standard deviation of Gauss convolver pdf.
const double Sigma=1.0;
// Width of convolver probability density function in case 
// of uniform and triangle probability density function.
const double Width=2.0;
// Flag to switch whether binning effect should be simulated.
const bool simulateBinning=true;
// Flag to switch transpose folding. Transpose folding is 
// the deconvolution for Pg*g*f, where f is the unknown and g is the original.
// convolver function.
const bool transposeFolding=true;
// Flag to switch smoothing or no smoothing. Smoothing is deconvolution 
// for h*g*f, where f is the unknown and g is the original convolver function, 
// while h is the smoother Gaussian.
const bool smoothing=false;
// Flag to switch on/off optimal scaling.
const double optimalScaling=true;
// Standard deviation of Gauss smoother pdf.
const double smootherSigma=(upperborder-lowerborder)/nbins*5;
// Smoothing for Bayesian method.
const double smoothing_bayesian=false;
// Maximum number of iterations.
const int maxIteration=40;
// Maximum number of iterations for Bayesian method.
const int maxIteration_bayesian=20;
// Maximum error tolerance for Bayesian method.
const double errorTolerance_bayesian=0.2;
// Calculate also inverse matrix norm?
const bool calculateInverseMatrixNorm=true;


// Parameters for the folding.
typedef struct
{
    // Minimum of histogram range.
    double lowerborderinput;
    // Maximum of histogram range.
    double upperborderinput;
    // Minimum of folded histogram range.
    double lowerborderfolded;
    // Maximum of folded histogram range.
    double upperborderfolded;
    // Type of convolver function.
    type Type;
    // Standard deviation of Gauss convolver pdf.
    double Sigma;
    // Width of convolver probability density function in case 
    // of uniform probability density function, and half-width in case of 
    // triangle probability density function.
    double Width;
    // With or without binning.
    bool withBinning;
}
parameters_t;


// The folding operator.
int folder(const int nbinsin, const double *inputpdf, const void *parameters, const int nbinsout, double *foldedpdf)
{
    const parameters_t *param=(const parameters_t*)parameters;
    const double Dx=(param->upperborderinput-param->lowerborderinput)/nbinsin;
    const double Dy=(param->upperborderfolded-param->lowerborderfolded)/nbinsout;
    const bool withBinning=param->withBinning;
    double params[3];
    if ( param->Type==Gauss )
    {
	params[0]=param->Sigma;
	params[1]=Dx;
	params[2]=Dy;
	if ( withBinning ) return convolveByGaussAndBinning1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
	else return convolveByGauss1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
    }
    else if ( param->Type==Uniform )
    {
	params[0]=param->Width;
	params[1]=Dx;
	params[2]=Dy;
	if ( withBinning ) return convolveByUniformAndBinning1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
	else return convolveByUniform1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
    }
    else if ( param->Type==Triangle )
    {
	params[0]=param->Width;
	params[1]=Dx;
	params[2]=Dy;
	if ( withBinning ) return convolveByTriangleAndBinning1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
	else return convolveByTriangle1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
    }
    fprintf(stderr, "[folder] Unknown probability density function.\n");
    return -1;
}




// The transpose folding operator.
int transposeFolder(const int nbinsin, const double *inputpdf, const void *parameters, const int nbinsout, double *foldedpdf)
{
    const parameters_t *param=(const parameters_t*)parameters;
    const double Dx=(param->upperborderinput-param->lowerborderinput)/nbinsout;
    const double Dy=(param->upperborderfolded-param->lowerborderfolded)/nbinsin;
    const bool withBinning=param->withBinning;
    double params[3];
    if ( param->Type==Gauss )
    {
	params[0]=param->Sigma;
	params[1]=Dx;
	params[2]=Dy;
	if ( withBinning ) return transposeConvolveByGaussAndBinning1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
	else { params[1]=Dy; params[2]=Dx; return convolveByGauss1D(nbinsin, inputpdf, params, nbinsout, foldedpdf); }
    }
    else if ( param->Type==Uniform )
    {
	params[0]=param->Width;
	params[1]=Dx;
	params[2]=Dy;
	if ( withBinning) return transposeConvolveByUniformAndBinning1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
	else { params[1]=Dy; params[2]=Dx; return convolveByUniform1D(nbinsin, inputpdf, params, nbinsout, foldedpdf); }
    }
    else if ( param->Type==Triangle )
    {
	params[0]=param->Width;
	params[1]=Dx;
	params[2]=Dy;
	if ( withBinning ) return transposeConvolveByTriangleAndBinning1D(nbinsin, inputpdf, params, nbinsout, foldedpdf);
	else { params[1]=Dy; params[2]=Dx; return convolveByTriangle1D(nbinsin, inputpdf, params, nbinsout, foldedpdf); }
    }
    fprintf(stderr, "[transposeFolder] Unknown probability density function.\n");
    return -1;
}




// Main program:
int main(int argc, char *argv[])
{
    // Set up parameters.
    const double lowerborderfolded=lowerborder;
    const double upperborderfolded=upperborder;
    const int nbinsfolded=nbins;
    const double binwidthfolded=(upperborderfolded-lowerborderfolded)/nbinsfolded;
    const double lowerborderunfolded=-0.5*(upperborder-lowerborder)*widthratio+0.5*(upperborder+lowerborder);
    const double upperborderunfolded=0.5*(upperborder-lowerborder)*widthratio+0.5*(upperborder+lowerborder);
    const int nbinsunfolded=nbins*widthratio*samplingratio;
    const double binwidthunfolded=binwidthfolded/samplingratio;
    const double lowerborderinput=lowerborderunfolded;
    const double upperborderinput=upperborderunfolded;
    const int nbinsinput=nbinsunfolded;
    const double binwidthinput=binwidthunfolded;
    const double smoothingGaussParameters[]={smootherSigma, binwidthfolded, binwidthfolded};
    parameters_t parameters;
    parameters.lowerborderinput=lowerborderinput;
    parameters.upperborderinput=upperborderinput;
    parameters.lowerborderfolded=lowerborderfolded;
    parameters.upperborderfolded=upperborderfolded;
    parameters.Type=Type;
    parameters.Sigma=Sigma;
    parameters.Width=Width;
    parameters.withBinning=simulateBinning;
    parameters_t parameters_naive;
    parameters_naive.lowerborderinput=lowerborderfolded;
    parameters_naive.upperborderinput=upperborderfolded;
    parameters_naive.lowerborderfolded=lowerborderfolded;
    parameters_naive.upperborderfolded=upperborderfolded;
    parameters_naive.Type=Type;
    parameters_naive.Sigma=Sigma;
    parameters_naive.Width=Width;
    parameters_naive.withBinning=simulateBinning;

    vector<double> inputpdf(nbinsinput);
    vector<double> inputpdferr(nbinsinput);
    vector<double> inputpdf_naive(nbinsfolded);
    vector<double> inputpdferr_naive(nbinsfolded);
    vector<double> foldedpdf(nbinsfolded);
    vector<double> foldedpdferr(nbinsfolded);
    vector<double*> foldedpdfErr(nbins, NULL);
    vector< vector<double> > foldedpdfErrBuff(nbins, vector<double>(nbins, 0.0));
    for ( int i=0 ; i<nbins ; ++i ) foldedpdfErr[i]=&foldedpdfErrBuff[i][0];
    vector<double> unfoldedpdf(nbinsunfolded);
    vector<double> unfoldedpdferr(nbinsunfolded);
    vector<double*> unfoldedpdfErr(nbinsunfolded, NULL);
    vector< vector<double> > unfoldedpdfErrBuff(nbinsunfolded, vector<double>(nbins, 0.0));
    for ( int i=0 ; i<nbinsunfolded ; ++i ) unfoldedpdfErr[i]=&unfoldedpdfErrBuff[i][0];
    vector<double> unfoldedpdf_naive(nbinsfolded);
    vector<double> unfoldedpdferr_naive(nbinsfolded);
    vector<double> unfoldedpdf_coarsegrained(nbinsfolded);
    vector<double> unfoldedpdferr_coarsegrained(nbinsfolded);
    vector<double> unfoldedpdfcnt_coarsegrained(nbinsfolded);
    memset(&inputpdf[0], 0, nbinsinput*sizeof(double));
    memset(&inputpdferr[0], 0, nbinsinput*sizeof(double));
    memset(&inputpdf_naive[0], 0, nbinsfolded*sizeof(double));
    memset(&inputpdferr_naive[0], 0, nbinsfolded*sizeof(double));
    memset(&foldedpdf[0], 0, nbinsfolded*sizeof(double));
    memset(&foldedpdferr[0], 0, nbinsfolded*sizeof(double));
    memset(&unfoldedpdf[0], 0, nbinsunfolded*sizeof(double));
    memset(&unfoldedpdferr[0], 0, nbinsunfolded*sizeof(double));
    memset(&unfoldedpdf_naive[0], 0, nbinsfolded*sizeof(double));
    memset(&unfoldedpdferr_naive[0], 0, nbinsfolded*sizeof(double));
    memset(&unfoldedpdf_coarsegrained[0], 0, nbinsfolded*sizeof(double));
    memset(&unfoldedpdferr_coarsegrained[0], 0, nbinsfolded*sizeof(double));
    memset(&unfoldedpdfcnt_coarsegrained[0], 0, nbinsfolded*sizeof(double));

    ofstream ofile;

    // Generate input pdf as a Cauchy distribution.
    cout<<"Generating input pdf.\n"; cout.flush();
    for ( int i=0 ; i<nbinsinput ; ++i )
    {
	const double x=lowerborderinput+binwidthinput*(i+0.5);
	inputpdf[i]=1.0/(M_PI*Gamma)*1.0/(1.0+(x*x/(Gamma*Gamma)));
	inputpdferr[i]=0.0;
    }
    for ( int i=0 ; i<nbinsfolded ; ++i )
    {
	const double x=lowerborderfolded+binwidthfolded*(i+0.5);
	inputpdf_naive[i]=1.0/(M_PI*Gamma)*1.0/(1.0+(x*x/(Gamma*Gamma)));
	inputpdferr_naive[i]=0.0;
    }

    // Fold the input pdf.
    cout<<"Generating folded pdf.\n"; cout.flush();
    if ( !withMC )
    {
	if ( folder(nbinsinput, &inputpdf[0], &parameters, nbinsfolded, &foldedpdf[0]) )
	{
	    cerr<<"[main] Error on folding.\n";
	}
	for ( int i=0 ; i<nbinsfolded ; ++i )
	{
	    foldedpdferr[i]=0.0;
	    for ( int j=0 ; j<nbinsfolded ; ++j ) foldedpdfErr[i][j]=0.0;
	}
    }
    else
    {
	for ( int entry=0 ; entry<Entries ; ++entry )
	{
	    const double xcauchy=cauchy(0.0, Gamma);
	    const double xfolder=(Type==Gauss ? gauss(0.0, Sigma) : (Type==Uniform ? uniform(-0.5*Width, 0.5*Width) : (Type==Triangle ? triang(0.0, Width) : 0.0)));
	    const double x=xcauchy+xfolder;
	    int i=(int)((x-lowerborderfolded)/binwidthfolded);
	    if ( i<0 ) {  } // Throw away underflow values.
	    else if ( i+1>nbinsfolded ) {  } // Throw away overflow values.
	    else { ++foldedpdf[i]; } // Fill histogram.
	}
	for ( int i=0 ; i<nbinsfolded ; ++i )
	{
	    foldedpdferr[i]=::sqrt(::fabs(foldedpdf[i]))/(binwidthfolded*Entries);
	    foldedpdf[i]/=(binwidthfolded*Entries);
	    for ( int j=0 ; j<nbinsfolded ; ++j ) foldedpdfErr[i][j]=0.0;
	    foldedpdfErr[i][i]=foldedpdferr[i];
	}
    }

    // Unfold by deconvolution.
    cout<<"Unfolding by deconvolution.\n"; cout.flush();
    parameters_naive.withBinning=false;
    if ( deconvolve_dft(nbinsfolded, &foldedpdf[0], folder, &parameters_naive, nbinsfolded, &unfoldedpdf_naive[0]) )
    {
	cerr<<"[main] Error on unfolding by dft.\n";
    }
    parameters_naive.withBinning=simulateBinning;
    ofile.open("out/data/unfoldedpdf_dft.dat");
    ofile<<"## This is the unfolded probability density function using DFT.\n";
    ofile<<"# x\t\tinputpdf(x)\t\tinputpederror(x)\t\tfoldedpdf(x)\t\tfoldedpdferror(x)\t\tunfoldedpdf(x)\t\tunfoldedpdferror(x)\n";
    for ( int i=0 ; i<nbinsfolded ; ++i )
    {
	const double x=lowerborderfolded+binwidthfolded*(i+0.5);
	ofile<<x<<"\t"<<inputpdf_naive[i]<<"\t"<<inputpdferr_naive[i]<<"\t"<<foldedpdf[i]<<"\t"<<foldedpdferr[i]<<"\t"<<unfoldedpdf_naive[i]<<"\t"<<unfoldedpdferr_naive[i]<<endl;
    }
    ofile.close();
    ofile.clear();

    // Unfold by inverse matrix.
    cout<<"Unfolding by inverse matrix.\n"; cout.flush();
    if ( unfold_inverse(nbinsfolded, &foldedpdf[0], folder, &parameters_naive, nbinsfolded, &unfoldedpdf_naive[0]) )
    {
	cerr<<"[main] Error on unfolding by inverse.\n";
    }
    ofile.open("out/data/unfoldedpdf_inverse.dat");
    ofile<<"## This is the unfolded probability density function using matrix inversion.\n";
    ofile<<"# x\t\tinputpdf(x)\t\tinputpederror(x)\t\tfoldedpdf(x)\t\tfoldedpdferror(x)\t\tunfoldedpdf(x)\t\tunfoldedpdferror(x)\n";
    for ( int i=0 ; i<nbinsfolded ; ++i )
    {
	const double x=lowerborderfolded+binwidthfolded*(i+0.5);
	ofile<<x<<"\t"<<inputpdf_naive[i]<<"\t"<<inputpdferr_naive[i]<<"\t"<<foldedpdf[i]<<"\t"<<foldedpdferr[i]<<"\t"<<unfoldedpdf_naive[i]<<"\t"<<unfoldedpdferr_naive[i]<<endl;
    }
    ofile.close();
    ofile.clear();

    // Determine operator norm of inverse matrix.
    if ( calculateInverseMatrixNorm )
    {

	cout<<"Calculating inverse matrix norm.\n"; cout.flush();
	double inverseOperatorNorm=0.0;
	if ( unfold_inverseoperatornorm(nbinsfolded, folder, &parameters_naive, nbinsfolded, &inverseOperatorNorm) )
	{
	    cerr<<"[main] Error on calculating inverse operator norm.\n";
	}
	cout<<"inverseOperatorNorm="<<inverseOperatorNorm<<endl; cout.flush();
    }

    // Unfold with Bayesian iteration method.
    cout<<"Unfolding by Bayesian iterative method.\n"; cout.flush();
    int nIterations_bayesian=0;
    double errorContent_bayesian=0.0;
    vector<double> dAlembertL1_bayesian(maxIteration_bayesian);
    vector<double> dAlembertL2_bayesian(maxIteration_bayesian);
    vector<double> dAlembertLi_bayesian(maxIteration_bayesian);
    if ( unfold_bayesian(nbinsfolded, &foldedpdf[0], NULL, &foldedpdferr[0], 
                         folder, &parameters_naive, 
                         (smoothing_bayesian ? convolveByGauss1D : NULL), smoothingGaussParameters, 
                         maxIteration_bayesian, errorTolerance_bayesian, 
                         nbinsfolded, &unfoldedpdf_naive[0], NULL, &unfoldedpdferr_naive[0], 
                         &nIterations_bayesian, &errorContent_bayesian, 
                         &dAlembertL1_bayesian[0], &dAlembertL2_bayesian[0], &dAlembertLi_bayesian[0]) )
    {
	cerr<<"[main] Error on unfolding by Bayesian method.\n";
    }
    ofile.open("out/data/unfoldedpdf_bayesian.dat");
    ofile<<"## This is the unfolded probability density function using Bayesian method.\n";
    ofile<<"# x\t\tinputpdf(x)\t\tinputpederror(x)\t\tfoldedpdf(x)\t\tfoldedpdferror(x)\t\tunfoldedpdf(x)\t\tunfoldedpdferror(x)\n";
    for ( int i=0 ; i<nbinsfolded ; ++i )
    {
	const double x=lowerborderfolded+binwidthfolded*(i+0.5);
	ofile<<x<<"\t"<<inputpdf_naive[i]<<"\t"<<inputpdferr_naive[i]<<"\t"<<foldedpdf[i]<<"\t"<<foldedpdferr[i]<<"\t"<<unfoldedpdf_naive[i]<<"\t"<<unfoldedpdferr_naive[i]<<endl;
    }
    ofile.close();
    ofile.clear();
    ofile.open("out/data/unfoldedpdf_bayesian_dalembert.dat");
    ofile<<"## This is the d'Alembert convergence test for Bayesian unfolding.\n";
    ofile<<"# order\t\td'AlembertL1(order)\t\td'AlembertL2(order)\t\tdAlembertLinfty(order)\n";
    for ( int n=0 ; n<nIterations_bayesian ; ++n )
    {
	ofile<<n<<"\t\t"<<dAlembertL1_bayesian[n]<<"\t\t"<<dAlembertL2_bayesian[n]<<"\t\t"<<dAlembertLi_bayesian[n]<<endl;
    }
    ofile.close();
    ofile.clear();
    ofile.open("out/data/unfoldedpdf_bayesian_report.dat");
    ofile<<"nIterations= "<<nIterations_bayesian<<endl;
    ofile<<"errorContent= "<<errorContent_bayesian<<endl;
    ofile<<"entries= "<<Entries<<endl;
    ofile.close();
    ofile.clear();

    // Determine optimal scaling factor for linear iteration.
    double scalingFactor=1.0;
    if ( optimalScaling )
    {
	cout<<"Calculating optimal scaling factor for iterative unfolding.\n"; cout.flush();
	if ( unfold_scalefactor(nbinsfolded, 
                                folder, &parameters, 
                                (smoothing ? convolveByGauss1D : NULL), smoothingGaussParameters, 
                                transposeFolder, &parameters, 
                                nbinsunfolded, 
                                &scalingFactor) )
	{
	    cerr<<"[main] Error on calculating scaling factor operator norm.\n";
	}
	cout<<"scalingFactor="<<scalingFactor<<endl; cout.flush();
    }

    // Unfold by linear iteration.
    cout<<"Unfolding by linear iterative method.\n"; cout.flush();
    int nIterations=0;
    vector<double> errorContent(maxIteration, 0.0);
    vector<double> dAlembertL1(maxIteration, 0.0);
    vector<double> dAlembertL2(maxIteration, 0.0);
    vector<double> dAlembertLi(maxIteration, 0.0);
    if ( unfold_iterative(nbinsfolded, &foldedpdf[0], (const double**)&foldedpdfErr[0], &foldedpdferr[0], NULL, 
                          folder, &parameters, 
                          (smoothing ? convolveByGauss1D : NULL), smoothingGaussParameters, 
                          (transposeFolding ? transposeFolder : NULL), &parameters, 
			  NULL, NULL, 
                          1.0/scalingFactor, 
                          maxIteration, maxIteration, 0, NULL, NULL, 
                          nbinsunfolded, &unfoldedpdf[0], &unfoldedpdfErr[0], &unfoldedpdferr[0], NULL, NULL, 
                          &nIterations, NULL, &errorContent[0], NULL, 
                          &dAlembertL1[0], &dAlembertL2[0], &dAlembertLi[0]) )
    {
	cerr<<"[main] Error on unfolding by iteration.\n";
    }
    ofile.open("out/data/unfoldedpdf_iterative.dat");
    ofile<<"## This is the unfolded probability density function using iterative unfolding.\n";
    ofile<<"# x\t\tinputpdf(x)\t\tinputpederror(x)\t\tfoldedpdf(x)\t\tfoldedpdferror(x)\t\tunfoldedpdf(x)\t\tunfoldedpdferror(x)\n";
    for ( int iunfolded=0 ; iunfolded<nbinsunfolded ; ++iunfolded )
    {
	const double x=lowerborderunfolded+binwidthunfolded*(iunfolded+0.5);
	
	int ifolded=(int)((x-lowerborderfolded)/binwidthfolded);
	unfoldedpdferr[iunfolded]=0.0;
	for ( int j=0 ; j<nbins ; ++j ) unfoldedpdferr[iunfolded]+=unfoldedpdfErr[iunfolded][j]*unfoldedpdfErr[iunfolded][j];
	unfoldedpdferr[iunfolded]=::sqrt(unfoldedpdferr[iunfolded]);
	if ( ifolded<0 ) {  }
	else if ( ifolded+1>nbinsfolded ) {  }
	else
	{
	    unfoldedpdf_coarsegrained[ifolded]+=unfoldedpdf[iunfolded];
	    unfoldedpdferr_coarsegrained[ifolded]+=unfoldedpdferr[iunfolded]*unfoldedpdferr[iunfolded];
	    ++unfoldedpdfcnt_coarsegrained[ifolded];
	}
    }
    for ( int ifolded=0 ; ifolded<nbinsfolded ; ++ifolded )
    {
	const double x=lowerborderfolded+binwidthfolded*(ifolded+0.5);
	unfoldedpdf_coarsegrained[ifolded]/=unfoldedpdfcnt_coarsegrained[ifolded];
	unfoldedpdferr_coarsegrained[ifolded]=::sqrt(unfoldedpdferr_coarsegrained[ifolded])/unfoldedpdfcnt_coarsegrained[ifolded];
	ofile<<x<<"\t"<<inputpdf_naive[ifolded]<<"\t"<<inputpdferr_naive[ifolded]<<"\t"<<foldedpdf[ifolded]<<"\t"<<foldedpdferr[ifolded]<<"\t"<<unfoldedpdf_coarsegrained[ifolded]<<"\t"<<unfoldedpdferr_coarsegrained[ifolded]<<endl;
    }
//    for ( int iunfolded=0 ; iunfolded<nbinsunfolded ; ++iunfolded )
//    {
//	const double x=lowerborderunfolded+binwidthunfolded*(iunfolded+0.5);
//	int ifolded=(int)((x-lowerborderfolded)/binwidthfolded);
//	if ( ifolded<0 ) ifolded=0;
//	else if ( ifolded+1>nbinsfolded ) ifolded=nbinsfolded-1;
//	const double EPSILON=binwidthunfolded*0.125;
//	const double xfolded=lowerborderfolded+binwidthfolded*(ifolded+0.5);
//	const double foldedpdfvalue=(::fabs(xfolded-x)<EPSILON ? foldedpdf[ifolded] : NAN);
//	const double foldedpdferrvalue=(::fabs(xfolded-x)<EPSILON ? foldedpdferr[ifolded] : NAN);
//	ofile<<x<<"\t"<<inputpdf[iunfolded]<<"\t"<<inputpdferr[iunfolded]<<"\t"<<foldedpdfvalue<<"\t"<<foldedpdferrvalue<<"\t"<<unfoldedpdf[iunfolded]<<"\t"<<unfoldedpdferr[iunfolded]<<endl;
//    }
    ofile.close();
    ofile.clear();
    ofile.open("out/data/unfoldedpdf_iterative_dalembert.dat");
    ofile<<"## This is the d'Alembert convergence test for iterative unfolding.\n";
    ofile<<"# order\t\td'AlembertL1(order)\t\td'AlembertL2(order)\t\tdAlembertLinfty(order)\n";
    for ( int n=0 ; n<nIterations ; ++n )
    {
	ofile<<n<<"\t\t"<<dAlembertL1[n]<<"\t\t"<<dAlembertL2[n]<<"\t\t"<<dAlembertLi[n]<<endl;
    }
    ofile.close();
    ofile.clear();
    ofile.open("out/data/unfoldedpdf_iterative_report.dat");
    ofile<<"nIterations= "<<nIterations<<endl;
    ofile<<"errorContent= "<<(nIterations>0 ? errorContent[nIterations-1] : 0.0)<<endl;
    ofile<<"entries= "<<Entries<<endl;
    ofile.close();
    ofile.clear();

    return 0;
}


///////////////
// unfold.cc //
///////////////
