//////////////////////////////////////////////////////////////////////////////
// unfold_example.cc  This is a demo for unfolding calorimeter response.    //
//////////////////////////////////////////////////////////////////////////////


#include <rnd.h>
#include <unfold_inverse.h>
#include <unfold_iterative.h>
#include <folding_examples.h>
#include <invert.h>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <vector>
#include <iostream>
#include <fstream>
using namespace std;


// Temperature parameter of spectrum.
const double T=0.145;
// Tsallis index of spectrum.
const double n=6.6;
// Maximum energy in the histogram.
const double maxE=5.0;
// Number of bins in the histogram.
const int nbinsE=50;
// Entries in the histogram.
const int Entries=10000;
// Calorimeter parameters. sigma(E)=a+sqrt(b*E)+c*E, E in GeV.
const double a=0.150;
const double b=0.847*0.847;
const double c=0.074;
// Do transpose folding?
const bool transposeFolding=true;
// Smooth with Gaussian?
const bool smoothing=false;
// Sigma of smoother.
const double smootherSigma=maxE/nbinsE*5;
// Maximum number of iterations.
const int maxIteration=1600;
// Minimum number of iterations.
const int miniteration=10;
// Maximum allowed propagated combined statistical, systematic and approximation error of unfolded pdf.
const double errorthreshold=0.07;
// Overall systematic error added to the simulation.
const double syst=0.05;
// Parameters for systematic distortion (inefficiency).
const double transitionE=maxE/5.0;
const double scaleE=maxE/5.0;


// Inefficiency function to simulate uncorrected systematic errors.
double inefficiency(const double maxIneff, const double transitionE, const double scaleE, const double measuredE)
{
    return 0.5*::fabs(1.0+::tanh((measuredE-transitionE)/scaleE))*maxIneff;
}


// Main program:
int main(int argc, char *argv[])
{
    const double minE=-maxE;
    const int nbins=nbinsE*2;
    const double binwidth=(maxE-minE)/nbins;
    const double minEunfolded=minE;
    const double maxEunfolded=maxE;
    const int nbinsunfolded=nbins;
    const double binwidthunfolded=(maxEunfolded-minEunfolded)/nbinsunfolded;
    const double parameters[]={a, b, c, binwidthunfolded, binwidth};
    const double smoothingGaussParameters[]={smootherSigma, binwidth, binwidth};
    const double stoppingParams[]={errorthreshold};

    vector<double> inputpdf(nbinsunfolded, 0.0);
    vector<double> inputpdferr(nbinsunfolded, 0.0);
    double summedEntries=0.0;
    vector<double> foldedpdf(nbins, 0.0);
    vector<double> foldedpdferr(nbins, 0.0);
    vector<double*> foldedpdfErr(nbins, NULL);
    vector< vector<double> > foldedpdfErrBuff(nbins, vector<double>(nbins, 0.0));
    for ( int i=0 ; i<nbins ; ++i ) foldedpdfErr[i]=&foldedpdfErrBuff[i][0];
    vector<double> foldedpdfsyst(nbins, 0.0);
    vector<double> unfoldedpdf(nbinsunfolded, 0.0);
    vector<double> unfoldedpdferr(nbinsunfolded, 0.0);
    vector<double*> unfoldedpdfErr(nbinsunfolded, NULL);
    vector< vector<double> > unfoldedpdfErrBuff(nbinsunfolded, vector<double>(nbins, 0.0));
    for ( int i=0 ; i<nbinsunfolded ; ++i ) unfoldedpdfErr[i]=&unfoldedpdfErrBuff[i][0];
    vector<double> unfoldedpdfappr(nbinsunfolded, 0.0);
    vector<double> unfoldedpdfsyst(nbinsunfolded, 0.0);

    ofstream ofile;

    // Generate input pdf as observed in CMS.
    cout<<"Generating input pdf.\n"; cout.flush();
    for ( int i=0 ; i<nbinsunfolded ; ++i )
    {
	const double E=minEunfolded+binwidthunfolded*(i+0.5);
	inputpdf[i]=(E>0.0 ? ::fabs(E)*(n-1.0)*(n-2.0)/((n*T)*(n*T))*::pow(1.0+::fabs(E)/(n*T), -n) : 0.0);
	inputpdferr[i]=0.0;
    }

    // Fold the input pdf with calorimeter response.
    cout<<"Generating folded pdf.\n"; cout.flush();
    summedEntries=Entries;
    summedEntries=0.0;
    for ( int entry=0 ; entry<Entries ; ++entry )
    {
	const double E=idtsallis(T, n);
	const double mean=E;
	const double sigma=a+::sqrt(b*E)+c*E;
	const double Efolded=gauss(mean, sigma);
	const double ineff=inefficiency(syst, transitionE, scaleE, Efolded);
	if ( uniform(0.0, 1.0)<ineff ) continue;
	int i=(int)((Efolded-minE)/binwidth);
	if ( i<0 ) {  } // Throw away underflow.
	else if ( i+1>nbins ) {  } // Throw away overflow.
	else { ++foldedpdf[i]; } // Fill histogram.
	++summedEntries; // Count entries.
    }
    for ( int i=0 ; i<nbins ; ++i )
    {
	foldedpdferr[i]=::sqrt(::fabs(foldedpdf[i]))/(binwidth*summedEntries);
	foldedpdf[i]/=(binwidth*summedEntries);
	for ( int j=0 ; j<nbins ; ++j ) foldedpdfErr[i][j]=0.0;
	foldedpdfErr[i][i]=foldedpdferr[i];
    }
    for ( int i=0 ; i<nbins ; ++i )
    {
	foldedpdfsyst[i]=foldedpdf[i]*syst;
    }

    // Calculate optimal scaling.
    double scalingFactor=1.0;
    cout<<"Calculating optimal scaling factor.\n"; cout.flush();
    if ( unfold_scalefactor(nbins, 
			    foldByCalorimeterAndBinning, parameters, 
			    (smoothing ? convolveByGauss1D : NULL), smoothingGaussParameters, 
			    transposeFoldByCalorimeterAndBinning, parameters, 
			    nbinsunfolded, 
			    &scalingFactor) )
    {
	cerr<<"[main] Error on calculating optimal scaling factor.\n";
    }
    cout<<"scalingFactor="<<scalingFactor<<endl; cout.flush();

    // Unfold by iteration.
    cout<<"Unfold by linear iterative method.\n"; cout.flush();
    int nIterations=0;
    vector<double> maxApprox(maxIteration, 0.0);
    vector<double> maxStat(maxIteration, 0.0);
    vector<double> maxSyst(maxIteration, 0.0);
    vector<double> dAlembertL1(maxIteration, 0.0);
    vector<double> dAlembertL2(maxIteration, 0.0);
    vector<double> dAlembertLi(maxIteration, 0.0);
    if ( unfold_iterative(nbins, &foldedpdf[0], (const double**)&foldedpdfErr[0], &foldedpdferr[0], &foldedpdfsyst[0], 
			  foldByCalorimeterAndBinning, parameters, 
			  (smoothing ? convolveByGauss1D : NULL), smoothingGaussParameters, 
			  (transposeFolding ? transposeFoldByCalorimeterAndBinning : NULL), parameters, 
			  NULL, NULL, 
			  1.0/scalingFactor, 
			  miniteration, maxIteration, 
			  1, stopbycombinederr, stoppingParams, 
			  nbinsunfolded, &unfoldedpdf[0], &unfoldedpdfErr[0], &unfoldedpdferr[0], &unfoldedpdfappr[0], &unfoldedpdfsyst[0], 
			  &nIterations, 
			  &maxApprox[0], &maxStat[0], &maxSyst[0], 
			  &dAlembertL1[0], &dAlembertL2[0], &dAlembertLi[0]) )
    {
	cerr<<"[main] Error on unfolding by iteration.\n";
    }

    ofile.open("out/data/cmshcal_iterative.dat");
    ofile<<"## This is the unfolded probability density function using iterative unfolding.\n";
    ofile<<"# x\t\tinputpdf(x)\t\tinputpederror(x)\t\tfoldedpdf(x)\t\tfoldedpdferror(x)\t\tfoldedpdfsyst(x)\t\tunfoldedpdf(x)\t\tunfoldedpdferror(x)\t\tunfoldedpdfappr(x)\t\tunfoldedpdfsyst(x)\n";
    for ( int i=0 ; i<nbins ; ++i )
    {
	const double E=minE+binwidth*(i+0.5);
	ofile<<E<<"\t"<<inputpdf[i]<<"\t"<<inputpdferr[i]<<"\t"<<foldedpdf[i]<<"\t"<<foldedpdferr[i]<<"\t"<<foldedpdfsyst[i]<<"\t"<<unfoldedpdf[i]<<"\t"<<unfoldedpdferr[i]<<"\t"<<unfoldedpdfappr[i]<<"\t"<<unfoldedpdfsyst[i]<<endl;
    }
    ofile.close();
    ofile.clear();

    ofile.open("out/data/cmshcal_errorcontent.dat");
    ofile<<"## This is the error content values for iterative unfolding.\n";
    ofile<<"# order\t\tmaxApproxError\tmaxStatError\tmaxSystError\n";
    for ( int n=0 ; n<nIterations ; ++n )
    {
	ofile<<n<<"\t\t"<<maxApprox[n]<<"\t\t"<<maxStat[n]<<"\t\t"<<maxSyst[n]<<endl;
    }
    ofile.close();
    ofile.clear();

    ofile.open("out/data/cmshcal_dalembert.dat");
    ofile<<"## This is the d'Alembert convergence test for iterative unfolding.\n";
    ofile<<"# order\t\td'AlembertL1(order)\t\td'AlembertL2(order)\t\td'AlembertLinfty(order)\n";
    for ( int n=0 ; n<nIterations ; ++n )
    {
	ofile<<n<<"\t\t"<<dAlembertL1[n]<<"\t\t"<<dAlembertL2[n]<<"\t\t"<<dAlembertLi[n]<<endl;
    }
    ofile.close();
    ofile.clear();

    ofile.open("out/data/cmshcal_report.dat");
    ofile<<"scalingFactor= "<<scalingFactor<<endl;
    ofile<<"nIterations= "<<nIterations<<endl;
    ofile<<"maxApproxErr= "<<(nIterations>0 ? maxApprox[nIterations-1] : 0.0)<<endl;
    ofile<<"maxStatErr= "<<(nIterations>0 ? maxStat[nIterations-1] : 0.0)<<endl;
    ofile<<"maxSystErr= "<<(nIterations>0 ? maxSyst[nIterations-1] : 0.0)<<endl;
    ofile<<"entries= "<<summedEntries<<endl;
    ofile.close();
    ofile.clear();

    return 0;
}


///////////////////////
// unfold_example.cc //
///////////////////////
