/**
 * unfold_iterative.c   This is a demo library for unfolding using 
 *                      iterative method of 
 *                      A.Laszlo: J.Phys A39 (2006) 13621
 *                      A.Laszlo: JPCS 368 (2012) 012043
 *                      (with some improvements).
 *                      Definitions and implementations in C.
 */




/* Types. */
#include <unfold_iterative.h>
/* Standard C library headers. */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>




/** Do debug printouts if DEBUG is defined, othervise ignore debug printouts. */
#ifdef DEBUG
#define dfprintf(...) fprintf(stderr, __VA_ARGS__)
#else
#define dfprintf(...)
#endif




/** Define dummy version string if version string is undefined. */
#ifndef VERSION
#define VERSION "unknown version"
#endif




/** Returns the version string of the library. */
const char* version(void)
{
    static const char versionstring[]=VERSION;
    return versionstring;
}




/** Sets verbosity level. */
static int verbositylevel = 1;
void verbosity(const int level)
{
    verbositylevel = level;
}




/** Allocate vector. */
static double* allocatevector(const int dim)
{
    double* result=(double*)calloc(dim, sizeof(double));
    return result;
}




/** Deallocate vector. */
static void freevector(double* v)
{
    free(v);
}




/** Allocate dimout x dimin (rows x cols) matrix. */
static double** allocatematrix(const int dimout, const int dimin)
{
    int i;
    double** result=(double**)calloc(dimout, sizeof(double*));
    for ( i=0 ; i<dimout ; ++i ) result[i]=(double*)calloc(dimin, sizeof(double));
    return result;
}




/** Deallocate dimout x dimin (rows x cols) matrix. */
static void freematrix(double** m, const int dimout)
{
    int i;
    for ( i=0 ; i<dimout ; ++i ) free(m[i]);
    free(m);
}




/** Set vector to zero. */
static void annulatevector(const int dim, double* vector)
{
    memset(vector, '\0', dim*sizeof(double));
}




/* Multiply in-place a vector by a scale factor. */
static void scalevector(const double scalefactor, const int dim, double* vector)
{
    int i;
    for ( i=0 ; i<dim ; ++i ) vector[i]*=scalefactor;
}




/** Copy a vector. */
static void copyvector(const int dim, const double* sourcevector, double* destvector)
{
    memcpy(destvector, sourcevector, dim*sizeof(double));
}




/* Multiply in-place a matrix by a scale factor. */
static void scalematrix(const double scalefactor, const int dimout, const int dimin, double** matrix)
{
    int i, j;
    for ( i=0 ; i<dimout ; ++i )
    {
	for ( j=0 ; j<dimin ; ++j ) matrix[i][j]*=scalefactor;
    }
}




/** Copy a dimout x dimint (rows x cols) matrix. */
static void copymatrix(const int dimout, const int dimin, const double** sourcematrix, double** destmatrix)
{
    int i;
    for ( i=0 ; i<dimout ; ++i ) memcpy(destmatrix[i], sourcematrix[i], dimin*sizeof(double));
}




/** Type for parameter to a composition of folding operators (max 4). */
typedef struct
{
    /* Number of compositions (max 4). */
    int ncomposition;
    /* List of consecutive folding operators. */
    folding folder[4];
    /* List of parameter pointers of the consecutive folding operators. */
    void* parameter[4];
    /* List of output bin numbers of the consecutive folding operators. */
    int nbinsout[4];
}
paramcompose_t;




/**
 * Folding operator composed of folding operators.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at nbinsin 
 *         		space point locations.
 * @param  parameters	Additional parameters passed to the folding operator 
 *         		(pointer to a paramcompose_t object).
 * @param  nbinsout	Total number of bins in the output histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values sampled at nbinsout locations.
 * @return errorcode	Nonzero on error.
 */
static int compose(const int nbinsin, const double* inputpdf, 
		   const void* parameters, 
		   const int nbinsout, double* foldedpdf)
{
    /* Error code. */
    int errorcode=0;
    /* Parameters. */
    const paramcompose_t *paramcompose=(const paramcompose_t*)parameters;
    const int ncomposition=paramcompose->ncomposition;
    /* Indices. */
    int n=0;
    /* Number of bins. */
    int nbinsinbuff=0;
    int nbinsoutbuff=0;
    int nmaxbins=0;
    /* Buffers. */
    double* buff[2]={NULL, NULL};
    double* buffpointer[2]={NULL, NULL};
    int inputbuff=0;

    /* If empty composition, return input at output. */
    if ( ncomposition==0 )
    {
	if ( nbinsout!=nbinsin ) return -1;
	copyvector(nbinsin, inputpdf, foldedpdf);
	return 0;
    }

    /* Get maximal necessary size of internal buffers. */
    for ( n=0 ; n<ncomposition-1 ; ++n )
    {
	if ( nmaxbins<paramcompose->nbinsout[n] ) nmaxbins=paramcompose->nbinsout[n];
    }

    /* Loop over compositions. */
    for ( n=0 ; n<ncomposition ; ++n )
    {
	/* If we are at the first, take input as input buffer. */
	if ( n==0 ) { nbinsinbuff=nbinsin; buffpointer[inputbuff]=(double*)inputpdf; }
	else { nbinsinbuff=paramcompose->nbinsout[n-1]; buffpointer[inputbuff]=buff[inputbuff]; }
	/* If we are at the last, take output as output buffer. */
	if ( n+1==ncomposition ) { nbinsoutbuff=nbinsout; buffpointer[!inputbuff]=foldedpdf; }
	else { nbinsoutbuff=paramcompose->nbinsout[n]; buffpointer[!inputbuff]=buff[!inputbuff]; }
	/* Allocate buffers when necessary. */
	if ( buffpointer[inputbuff]==NULL ) { buff[inputbuff]=allocatevector(nmaxbins); buffpointer[inputbuff]=buff[inputbuff]; }
	if ( buffpointer[!inputbuff]==NULL ) { buff[!inputbuff]=allocatevector(nmaxbins); buffpointer[!inputbuff]=buff[!inputbuff]; }
	/* Fold successively. */
	if ( (paramcompose->folder[n])(nbinsinbuff, buffpointer[inputbuff], paramcompose->parameter[n], nbinsoutbuff, buffpointer[!inputbuff]) )
	{
	    dfprintf("[compose] Error on %d-th folding composition.\n", n);
	    errorcode=-1;
	    break;
	}
	/* Swap input/output buffers. */
	inputbuff=!inputbuff;
    }

    /* Free buffers. */
    if ( buff[inputbuff]!=NULL ) { freevector(buff[inputbuff]); buff[inputbuff]=NULL; }
    if ( buff[!inputbuff]!=NULL ) { freevector(buff[!inputbuff]); buff[!inputbuff]=NULL; }

    return errorcode;
}




/**
 * Apply folding operator to a matrix as row vector of histograms 
 * (which are the column vectors) and assign it to an other matrix.
 * 
 * @param  nrowsin	Total number of rows of the input matrix.
 * @param  ncolsin	Total number of columns of the input matrix.
 * @param  inputmat	Input matrix.
 * @param  foldingop	Folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  nrowsout	Total number of rows of the output matrix.
 * @param  ncolsout	Total number of columns of the output matrix.
 * @param  foldedmat	Buffer to return output matrix.
 * @return errorcode	Nonzero on error (e.g. if ncolsout!=ncolsin).
 */
static int applytomatrix(const int nrowsin, const int ncolsin, const double** inputmat, 
			 const folding foldingop, const void* parameters, 
			 const int nrowsout, const int ncolsout, double** foldedmat)
{
    /* Error code. */
    int errorcode=0;
    /* Indices. */
    int i=0, j=0;
    /* Buffers. */
    double* sampler=NULL;
    double* samplerresult=NULL;

    if ( ncolsout!=ncolsin )
    {
	dfprintf("[applytomatrix] ncolsout(%d)!=ncolsin(%d).\n", ncolsout, ncolsin);
	errorcode=-1;
	return errorcode;
    }

    /* Allocate memory. */
    sampler=allocatevector(nrowsin);
    samplerresult=allocatevector(nrowsout);

    /* Apply folding to each column. */
    for ( j=0 ; j<ncolsin ; ++j )
    {
	for ( i=0 ; i<nrowsin ; ++i ) sampler[i]=inputmat[i][j];
	if ( foldingop(nrowsin, sampler, parameters, nrowsout, samplerresult) )
	{
	    dfprintf("[applytomatrix] Failed to apply folding to %d-th column of matrix.\n", j);
	    errorcode=-2;
	    break;
	}
	for ( i=0 ; i<nrowsout ; ++i ) foldedmat[i][j]=samplerresult[i];
    }

    /* Deallocate memory. */
    freevector(sampler); sampler=NULL;
    freevector(samplerresult); samplerresult=NULL;

    return errorcode;
}




/**
 * Sample folding operator as a matrix.
 * 
 * @param  foldingop	Folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  nrows	Sampling of the output probability base space.
 * @param  ncols	Sampling of the input probability base space.
 * @param  matrix	Buffer to return output matrix (nbinsout x nbinsin).
 * @return errorcode	Nonzero on error.
 */
static int sampleasmatrix(const folding foldingop, const void* parameters, 
			  const int nrows, const int ncols, 
			  double** matrix)
{
    /* Error code. */
    int errorcode=0;
    /* Indices. */
    int i=0, j=0;
    /* Temporary vector for sampling the folding operator. */
    double* sampler=NULL;
    double* samplerresult=NULL;

    /* Initialize folder matrix. */
    sampler=allocatevector(ncols);
    samplerresult=allocatevector(nrows);
    annulatevector(ncols, sampler);
    for ( j=0 ; j<ncols ; ++j )
    {
	if ( j>0 ) sampler[j-1]=0.0;
	sampler[j]=1.0;
	if ( foldingop(ncols, sampler, parameters, nrows, samplerresult) )
	{
	    dfprintf("[sampleasmatrix] Error on folding column %d.\n", j);
	    errorcode=-1;
	    break;
	}
	for ( i=0 ; i<nrows ; ++i ) matrix[i][j]=samplerresult[i];
    }
    freevector(samplerresult); samplerresult=NULL;
    freevector(sampler); sampler=NULL;

    return errorcode;
}




/** Macro for square function. */
#define square(x) ((x)*(x))




/** Macro for min. */
#define min(x,y) ((x)<(y) ? (x) : (y))




/**
 * The simplest iteration stopping function: stop if approximation error is 
 * smaller than a threshold.
 * (see: header file for meaning of parameters).
 */
int stopatgoodapprox(const double maxapprox,
		     const double maxstat,
		     const double maxsyst,
		     const void* threshold)
{
    const double approxthreshold=((const double*)threshold)[0];
    return (maxapprox<approxthreshold);
}




/**
 * Other simple iteration stopping function: stop when propagated.
 * statistical error exceeds approximation error.
 * (see: header file for meaning of parameters).
 */
int stopbystat(const double maxapprox,
	       const double maxstat,
	       const double maxsyst,
	       const void* null)
{
    return (maxstat>maxapprox);
}




/**
 * Other simple iteration stopping function: stop when propagated.
 * systematic error exceeds a threshold.
 * (see: header file for meaning of parameters).
 */
int stopbysyst(const double maxapprox,
	       const double maxstat,
	       const double maxsyst,
	       const void* threshold)
{
    const double systthreshold=((const double*)threshold)[0];
    return (maxsyst>systthreshold);
}




/**
 * Other simple iteration stopping function: stop when combined 
 * propagated statistical and systematic error exceeds a threshold.
 * (see: header file for meaning of parameters).
 */
int stopbystatandsyst(const double maxapprox,
		      const double maxstat,
		      const double maxsyst,
		      const void* threshold)
{
    const double statandsystthreshold=((const double*)threshold)[0];
    return (sqrt(square(maxstat)+square(maxsyst))>statandsystthreshold);
}




/**
 * Other simple iteration stopping function: stop when combined 
 * propagated statistical, systematic and approximation error exceeds a threshold.
 * (see: header file for meaning of parameters).
 */
int stopbycombinederr(const double maxapprox,
		      const double maxstat,
		      const double maxsyst,
		      const void* threshold)
{
    const double combinedthreshold=((const double*)threshold)[0];
    return (sqrt(square(maxapprox)+square(maxstat)+square(maxsyst))>combinedthreshold);
}




/**
 * Procedureto calculate a specific operator norm of the folding operator 
 * to judge convergence in any dimensional probability space 
 * (see: header file for meaning of parameters).
 */
int unfold_scalefactor(const int nbinsin, 
		       const folding foldingop, const void* parameters, 
		       const folding foldingopS, const void* parametersS, 
		       const folding foldingopT, const void* parametersT, 
		       const int nbinsout, 
		       double* operatornorm)
{
    /* Indices. */
    int i=0, j=0;
    /* Composition parameters. */
    paramcompose_t paramcompose;
    /* Folder matrix. */
    double** matrix=NULL;
    /* Matrix norm. */
    double matrixnorm=0.0;

    /* Set operatornorm to zero. */
    *operatornorm=0.0;

    /* Initialize double folder matrix. */
    paramcompose.ncomposition=0;
    if ( foldingop!=NULL )
    {
	paramcompose.folder[paramcompose.ncomposition]=foldingop;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parameters;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsin;
	++paramcompose.ncomposition;
    }
    if ( foldingopS!=NULL )
    {
	paramcompose.folder[paramcompose.ncomposition]=foldingopS;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parametersS;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsin;
	++paramcompose.ncomposition;
    }
    if ( foldingopT!=NULL )
    {
	paramcompose.folder[paramcompose.ncomposition]=foldingopT;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parametersT;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsout;
	++paramcompose.ncomposition;
    }
    matrix=allocatematrix(nbinsout, nbinsout);
    if ( sampleasmatrix(compose, &paramcompose, nbinsout, nbinsout, matrix) )
    {
	dfprintf("[unfold_scalefactor] Failed to sample as matrix.\n");
	freematrix(matrix, nbinsout); matrix=NULL;
	return -1;
    }

    /* Calculate L^infinity-induced matrix norm of folder matrix. */
    matrixnorm=0.0;
    for ( i=0 ; i<nbinsout ; ++i )
    {
	double l1normrow=0.0;
	for ( j=0 ; j<nbinsout ; ++j )
	{
	    l1normrow+=fabs(matrix[i][j]);
	}
	if ( matrixnorm<l1normrow ) matrixnorm=l1normrow;
    }

    /* Deallocate buffer memories. */
    freematrix(matrix, nbinsout); matrix=NULL;

    /* Report result. */
    *operatornorm=matrixnorm;

    return 0;
}




/**
 * Procedurefor unfolding using iteration in any dimensional 
 * probability space (see: header file for meaning of parameters).
 */
int unfold_iterative(const int nbinsin, const double* inputpdf, 
		     const double** inputErr, const double* inputerr, const double* inputsys, 
		     const folding foldingop, const void* parameters, 
		     const folding foldingopS, const void* parametersS, 
		     const folding foldingopT, const void* parametersT, 
		     const folding foldopsysT, const void* paramssysT, 
		     const double scalefactor, 
		     const int miniteration, const int maxiteration, 
		     const int calcappr, 
		     const stoppingfunc stopping, const void* stopparams, 
		     const int nbinsout, double* unfoldedpdf, 
		     double** unfoldedErr, double* unfoldederr, 
		     double* unfoldedappr, double* unfoldedsys, 
		     int *niterations, 
		     double* maxappr, double* maxstat, double* maxsyst, 
		     double* dAlembertL1, double* dAlembertL2, double* dAlembertLi)
{
    /* Error code. */
    int errorcode=0;
    /* Indices. */
    int i=0, j=0, n=0;
    /* L^1 norm of input pdf. */
    double inputpdfL1=0.0;
    /* Initial composition parameters. */
    paramcompose_t paramcomposeinitial;
    /* Initial composition parameters transposed. */
    paramcompose_t paramcomposeinitialtranspose;
    /* Composition parameters during iteration. */
    paramcompose_t paramcompose;
    /* d'Alembert convergence factor components. */
    double dAlembertl1=1.0;
    double dAlembertl1prev=1.0;
    double dAlembertl2=1.0;
    double dAlembertl2prev=1.0;
    double dAlembertli=1.0;
    double dAlembertliprev=1.0;
    double incr=0.0;
    /* Buffers for unfolded pdf. */
    double* f0=NULL;
    double* fn=NULL;
    double* fr=NULL;
    /* Buffers for propagated error matrix of unfolded pdf. */
    double** Err0=NULL;
    double** Errn=NULL;
    double** Errr=NULL;
    /* Buffers for estimated variance of unfolded pdf. */
    double* e0=NULL;
    double* en=NULL;
    double* ef=NULL;
    double** ma=NULL;
    double unfoldederri=0.0;
    /* Buffers for approximation error calculation for unfolded pdf. */
    double** xi=NULL;
    double** xi0=NULL;
    double** xin=NULL;
    double** xir=NULL;
    double* a=NULL;
    double xidiffdotunfolded=0.0;
    /* Buffers for propagated systematic error calculation for unfolded pdf. */
    double** Xi0=NULL;
    double** Xin=NULL;
    double** Xir=NULL;
    double** Ximod=NULL;
    double** Ximod2=NULL;
    double* s=NULL;
    double Ximoddotinputsys=0.0;
    double Ximod2Li=0.0;
    /* Maximum of bin probability errors for approximation, statistical and systematic errors. */
    double apprLi=0.0;
    double statLi=0.0;
    double systLi=0.0;

    if ( verbositylevel>0 )
    {
	fprintf(stdout, "INFO[unfold_iterative]: Initialization begins.\n"); fflush(stdout);
    }

    /* Calculate L^1 norm of initial pdf. */
    inputpdfL1=0.0;
    for ( i=0 ; i<nbinsin ; ++i )
    {
	inputpdfL1+=fabs(inputpdf[i]);
    }

    /* Prepare composite operators. */
    paramcomposeinitial.ncomposition=0;
    paramcomposeinitialtranspose.ncomposition=0;
    paramcompose.ncomposition=0;
    if ( foldingop!=NULL )
    {
	paramcomposeinitialtranspose.folder[paramcomposeinitialtranspose.ncomposition]=foldingop;
	paramcomposeinitialtranspose.parameter[paramcomposeinitialtranspose.ncomposition]=(void*)parameters;
	paramcomposeinitialtranspose.nbinsout[paramcomposeinitialtranspose.ncomposition]=nbinsin;
	++paramcomposeinitialtranspose.ncomposition;
	paramcompose.folder[paramcompose.ncomposition]=foldingop;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parameters;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsin;
	++paramcompose.ncomposition;
    }
    if ( foldingopS!=NULL )
    {
	paramcomposeinitialtranspose.folder[paramcomposeinitialtranspose.ncomposition]=foldingopS;
	paramcomposeinitialtranspose.parameter[paramcomposeinitialtranspose.ncomposition]=(void*)parametersS;
	paramcomposeinitialtranspose.nbinsout[paramcomposeinitialtranspose.ncomposition]=nbinsin;
	++paramcomposeinitialtranspose.ncomposition;
	paramcomposeinitial.folder[paramcomposeinitial.ncomposition]=foldingopS;
	paramcomposeinitial.parameter[paramcomposeinitial.ncomposition]=(void*)parametersS;
	paramcomposeinitial.nbinsout[paramcomposeinitial.ncomposition]=nbinsin;
	++paramcomposeinitial.ncomposition;
	paramcompose.folder[paramcompose.ncomposition]=foldingopS;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parametersS;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsin;
	++paramcompose.ncomposition;
    }
    if ( foldingopT!=NULL )
    {
	paramcomposeinitial.folder[paramcomposeinitial.ncomposition]=foldingopT;
	paramcomposeinitial.parameter[paramcomposeinitial.ncomposition]=(void*)parametersT;
	paramcomposeinitial.nbinsout[paramcomposeinitial.ncomposition]=nbinsout;
	++paramcomposeinitial.ncomposition;
	paramcompose.folder[paramcompose.ncomposition]=foldingopT;
	paramcompose.parameter[paramcompose.ncomposition]=(void*)parametersT;
	paramcompose.nbinsout[paramcompose.ncomposition]=nbinsout;
	++paramcompose.ncomposition;
    }

    /* Prepare initial value. */
    /* For the pdf. */
    f0=allocatevector(nbinsout);
    fn=allocatevector(nbinsout);
    if ( verbositylevel>0 )
    {
	fprintf(stdout, "INFO[unfold_iterative]: Initialization of pdf unfolding.\n"); fflush(stdout);
    }
    if ( compose(nbinsin, inputpdf, &paramcomposeinitial, nbinsout, f0) )
    {
	dfprintf("[unfold_iterative] Failed to prepare initial value.\n");
	if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	if ( fr!=NULL ) { freevector(fr); fr=NULL; }
	if ( Err0!=NULL ) { freematrix(Err0, nbinsout); Err0=NULL; }
	if ( Errn!=NULL ) { freematrix(Errn, nbinsout); Errn=NULL; }
	if ( Errr!=NULL ) { freematrix(Errr, nbinsout); Errr=NULL; }
	if ( e0!=NULL ) { freevector(e0); e0=NULL; }
	if ( en!=NULL ) { freevector(en); en=NULL; }
	if ( ef!=NULL ) { freevector(ef); ef=NULL; }
	if ( ma!=NULL ) { freematrix(ma, nbinsout); ma=NULL; }
	if ( xi!=NULL ) { freematrix(xi, nbinsout); xi=NULL; }
	if ( xi0!=NULL ) { freematrix(xi0, nbinsout); xi0=NULL; }
	if ( xin!=NULL ) { freematrix(xin, nbinsout); xin=NULL; }
	if ( xir!=NULL ) { freematrix(xir, nbinsout); xir=NULL; }
	if ( a!=NULL ) { freevector(a); a=NULL; }
	if ( Xi0!=NULL ) { freematrix(Xi0, nbinsout); Xi0=NULL; }
	if ( Xin!=NULL ) { freematrix(Xin, nbinsout); Xin=NULL; }
	if ( Xir!=NULL ) { freematrix(Xir, nbinsout); Xir=NULL; }
	if ( Ximod!=NULL ) { freematrix(Ximod, nbinsin); Ximod=NULL; }
	if ( Ximod2!=NULL ) { freematrix(Ximod2, nbinsout); Ximod2=NULL; }
	if ( s!=NULL ) { freevector(s); s=NULL; }
	return -2;
    }
    scalevector(scalefactor, nbinsout, f0);
    copyvector(nbinsout, f0, fn);
    /* For the statistical errors. */
    if ( inputErr!=NULL )
    {
	if ( verbositylevel>0 )
	{
	    fprintf(stdout, "INFO[unfold_iterative]: Initialization of statistical error propagation.\n"); fflush(stdout);
	}
	Err0=allocatematrix(nbinsout, nbinsin);
	Errn=allocatematrix(nbinsout, nbinsin);
	if ( applytomatrix(nbinsin, nbinsin, inputErr, compose, &paramcomposeinitial, nbinsout, nbinsin, Err0) )
	{
	    dfprintf("[unfold_iterative] Failed to prepare initial error matrix.\n");
	    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	    if ( fr!=NULL ) { freevector(fr); fr=NULL; }
	    if ( Err0!=NULL ) { freematrix(Err0, nbinsout); Err0=NULL; }
	    if ( Errn!=NULL ) { freematrix(Errn, nbinsout); Errn=NULL; }
	    if ( Errr!=NULL ) { freematrix(Errr, nbinsout); Errr=NULL; }
	    if ( e0!=NULL ) { freevector(e0); e0=NULL; }
	    if ( en!=NULL ) { freevector(en); en=NULL; }
	    if ( ef!=NULL ) { freevector(ef); ef=NULL; }
	    if ( ma!=NULL ) { freematrix(ma, nbinsout); ma=NULL; }
	    if ( xi!=NULL ) { freematrix(xi, nbinsout); xi=NULL; }
	    if ( xi0!=NULL ) { freematrix(xi0, nbinsout); xi0=NULL; }
	    if ( xin!=NULL ) { freematrix(xin, nbinsout); xin=NULL; }
	    if ( xir!=NULL ) { freematrix(xir, nbinsout); xir=NULL; }
	    if ( a!=NULL ) { freevector(a); a=NULL; }
	    if ( Xi0!=NULL ) { freematrix(Xi0, nbinsout); Xi0=NULL; }
	    if ( Xin!=NULL ) { freematrix(Xin, nbinsout); Xin=NULL; }
	    if ( Xir!=NULL ) { freematrix(Xir, nbinsout); Xir=NULL; }
	    if ( Ximod!=NULL ) { freematrix(Ximod, nbinsin); Ximod=NULL; }
	    if ( Ximod2!=NULL ) { freematrix(Ximod2, nbinsout); Ximod2=NULL; }
	    if ( s!=NULL ) { freevector(s); s=NULL; }
	    return -3;
	}
	scalematrix(scalefactor, nbinsout, nbinsin, Err0);
	copymatrix(nbinsout, nbinsin, (const double**)Err0, Errn);
    }
    else if ( inputerr!=NULL )
    {
	if ( verbositylevel>0 )
	{
	    fprintf(stdout, "INFO[unfold_iterative]: Initialization of statistical error propagation with neglected correlations.\n"); fflush(stdout);
	}
	e0=allocatevector(nbinsout);
	en=allocatevector(nbinsout);
	ma=allocatematrix(nbinsout, nbinsin);
	if ( sampleasmatrix(compose, &paramcomposeinitial, nbinsout, nbinsin, ma) )
	{
	    dfprintf("[unfold_iterative] Failed to sample as matrix initially.\n");
	    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	    if ( fr!=NULL ) { freevector(fr); fr=NULL; }
	    if ( Err0!=NULL ) { freematrix(Err0, nbinsout); Err0=NULL; }
	    if ( Errn!=NULL ) { freematrix(Errn, nbinsout); Errn=NULL; }
	    if ( Errr!=NULL ) { freematrix(Errr, nbinsout); Errr=NULL; }
	    if ( e0!=NULL ) { freevector(e0); e0=NULL; }
	    if ( en!=NULL ) { freevector(en); en=NULL; }
	    if ( ef!=NULL ) { freevector(ef); ef=NULL; }
	    if ( ma!=NULL ) { freematrix(ma, nbinsout); ma=NULL; }
	    if ( xi!=NULL ) { freematrix(xi, nbinsout); xi=NULL; }
	    if ( xi0!=NULL ) { freematrix(xi0, nbinsout); xi0=NULL; }
	    if ( xin!=NULL ) { freematrix(xin, nbinsout); xin=NULL; }
	    if ( xir!=NULL ) { freematrix(xir, nbinsout); xir=NULL; }
	    if ( a!=NULL ) { freevector(a); a=NULL; }
	    if ( Xi0!=NULL ) { freematrix(Xi0, nbinsout); Xi0=NULL; }
	    if ( Xin!=NULL ) { freematrix(Xin, nbinsout); Xin=NULL; }
	    if ( Xir!=NULL ) { freematrix(Xir, nbinsout); Xir=NULL; }
	    if ( Ximod!=NULL ) { freematrix(Ximod, nbinsin); Ximod=NULL; }
	    if ( Ximod2!=NULL ) { freematrix(Ximod2, nbinsout); Ximod2=NULL; }
	    if ( s!=NULL ) { freevector(s); s=NULL; }
	    return -4;
	}
	for ( i=0 ; i<nbinsout ; ++i )
	{
	    e0[i]=0.0;
	    for ( j=0 ; j<nbinsin ; ++j ) e0[i]+=square(scalefactor*ma[i][j]*inputerr[j]);
	}
	freematrix(ma, nbinsout); ma=NULL;
	copyvector(nbinsout, e0, en);
    }
    /* For the approximation errors. */
    if ( calcappr )
    {
	if ( verbositylevel>0 )
	{
	    fprintf(stdout, "INFO[unfold_iterative]: Initialization of approximation error propagation.\n"); fflush(stdout);
	}
	xi=allocatematrix(nbinsout, nbinsout);
	xi0=allocatematrix(nbinsout, nbinsout);
	xin=allocatematrix(nbinsout, nbinsout);
	for ( i=0 ; i<nbinsout ; ++i )
	{
	    for ( j=0 ; j<nbinsout ; ++j ) xi[i][j]=(i==j ? 1.0 : 0.0);
	}
	if ( applytomatrix(nbinsout, nbinsout, (const double**)xi, compose, &paramcompose, nbinsout, nbinsout, xi0) )
	{
	    dfprintf("[unfold_iterative] Failed to prepare initial xi0 matrix.\n");
	    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
	    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
	    if ( fr!=NULL ) { freevector(fr); fr=NULL; }
	    if ( Err0!=NULL ) { freematrix(Err0, nbinsout); Err0=NULL; }
	    if ( Errn!=NULL ) { freematrix(Errn, nbinsout); Errn=NULL; }
	    if ( Errr!=NULL ) { freematrix(Errr, nbinsout); Errr=NULL; }
	    if ( e0!=NULL ) { freevector(e0); e0=NULL; }
	    if ( en!=NULL ) { freevector(en); en=NULL; }
	    if ( ef!=NULL ) { freevector(ef); ef=NULL; }
	    if ( ma!=NULL ) { freematrix(ma, nbinsout); ma=NULL; }
	    if ( xi!=NULL ) { freematrix(xi, nbinsout); xi=NULL; }
	    if ( xi0!=NULL ) { freematrix(xi0, nbinsout); xi0=NULL; }
	    if ( xin!=NULL ) { freematrix(xin, nbinsout); xin=NULL; }
	    if ( xir!=NULL ) { freematrix(xir, nbinsout); xir=NULL; }
	    if ( a!=NULL ) { freevector(a); a=NULL; }
	    if ( Xi0!=NULL ) { freematrix(Xi0, nbinsout); Xi0=NULL; }
	    if ( Xin!=NULL ) { freematrix(Xin, nbinsout); Xin=NULL; }
	    if ( Xir!=NULL ) { freematrix(Xir, nbinsout); Xir=NULL; }
	    if ( Ximod!=NULL ) { freematrix(Ximod, nbinsin); Ximod=NULL; }
	    if ( Ximod2!=NULL ) { freematrix(Ximod2, nbinsout); Ximod2=NULL; }
	    if ( s!=NULL ) { freevector(s); s=NULL; }
	    return -5;
	}
	scalematrix(scalefactor, nbinsout, nbinsout, xi0);
	copymatrix(nbinsout, nbinsout, (const double **)xi0, xin);
	a=allocatevector(nbinsout);
	annulatevector(nbinsout, a);
    }
    /* For the systematic errors. */
    if ( inputsys!=NULL || foldopsysT!=NULL )
    {
	Xi0=allocatematrix(nbinsout, nbinsout);
	Xin=allocatematrix(nbinsout, nbinsout);
	for ( i=0 ; i<nbinsout ; ++i )
	{
	    for ( j=0 ; j<nbinsout ; ++j ) Xi0[i][j]=(i==j ? 1.0 : 0.0);
	}
	copymatrix(nbinsout, nbinsout, (const double **)Xi0, Xin);
	s=allocatevector(nbinsout);
	annulatevector(nbinsout, s);
    }

    /* If iteration is going to be performed, do some further initialization. */
    if ( maxiteration>0 )
    {
	/* For the pdf. */
	fr=allocatevector(nbinsout);
	/* For the statistical errors. */
	if ( inputErr!=NULL ) Errr=allocatematrix(nbinsout, nbinsin);
	else if ( inputerr!=NULL )
	{
	    if ( verbositylevel>0 )
	    {
		fprintf(stdout, "INFO[unfold_iterative]: Initialization of statistical error buffers.\n"); fflush(stdout);
	    }
	    ef=allocatevector(nbinsout);
	    ma=allocatematrix(nbinsout, nbinsout);
	    if ( sampleasmatrix(compose, &paramcompose, nbinsout, nbinsout, ma) )
	    {
		dfprintf("[unfold_iterative] Failed to sample as matrix.\n");
		if ( f0!=NULL ) { freevector(f0); f0=NULL; }
		if ( fn!=NULL ) { freevector(fn); fn=NULL; }
		if ( fr!=NULL ) { freevector(fr); fr=NULL; }
		if ( Err0!=NULL ) { freematrix(Err0, nbinsout); Err0=NULL; }
		if ( Errn!=NULL ) { freematrix(Errn, nbinsout); Errn=NULL; }
		if ( Errr!=NULL ) { freematrix(Errr, nbinsout); Errr=NULL; }
		if ( e0!=NULL ) { freevector(e0); e0=NULL; }
		if ( en!=NULL ) { freevector(en); en=NULL; }
		if ( ef!=NULL ) { freevector(ef); ef=NULL; }
		if ( ma!=NULL ) { freematrix(ma, nbinsout); ma=NULL; }
		if ( xi!=NULL ) { freematrix(xi, nbinsout); xi=NULL; }
		if ( xi0!=NULL ) { freematrix(xi0, nbinsout); xi0=NULL; }
		if ( xin!=NULL ) { freematrix(xin, nbinsout); xin=NULL; }
		if ( xir!=NULL ) { freematrix(xir, nbinsout); xir=NULL; }
		if ( a!=NULL ) { freevector(a); a=NULL; }
		if ( Xi0!=NULL ) { freematrix(Xi0, nbinsout); Xi0=NULL; }
		if ( Xin!=NULL ) { freematrix(Xin, nbinsout); Xin=NULL; }
		if ( Xir!=NULL ) { freematrix(Xir, nbinsout); Xir=NULL; }
		if ( Ximod!=NULL ) { freematrix(Ximod, nbinsin); Ximod=NULL; }
		if ( Ximod2!=NULL ) { freematrix(Ximod2, nbinsout); Ximod2=NULL; }
		if ( s!=NULL ) { freevector(s); s=NULL; }
		return -6;
	    }
	}
	/* For the approximation errors. */
	if ( calcappr )
	{
	     xir=allocatematrix(nbinsout, nbinsout);
	}
	/* For the systematic errors. */
	if ( inputsys!=NULL || foldopsysT!=NULL )
	{
	     Xir=allocatematrix(nbinsout, nbinsout);
	     Ximod=allocatematrix(nbinsin, nbinsout);
	     if ( foldopsysT!=NULL ) Ximod2=allocatematrix(nbinsout, nbinsout);
	}
    }

    if ( verbositylevel>0 )
    {
	fprintf(stdout, "INFO[unfold_iterative]: Initialization ends.\n"); fflush(stdout);
    }

    /* Perform the iteration. */
    dAlembertl1=1.0;
    dAlembertl1prev=1.0;
    dAlembertl2=1.0;
    dAlembertl2prev=1.0;
    dAlembertli=1.0;
    dAlembertliprev=1.0;
    for ( n=0 ; n<maxiteration ; ++n )
    {
	if ( verbositylevel>0 )
	{
	    fprintf(stdout, "INFO[unfold_iterative]: Iteration step %d begins.\n", n+1); fflush(stdout);
	}
	/* Calculate the n-th iteration. */
	/* Do it for the pdf and statistical errors. */
	if ( compose(nbinsout, fn, &paramcompose, nbinsout, fr) )
	{
	    dfprintf("[unfold_iterative] Failed in iteration step %d.\n", n+1);
	    errorcode=-7;
	    break;
	}
	if ( inputErr==NULL && inputerr!=NULL )
	{
	    for ( i=0 ; i<nbinsout ; ++i )
	    {
		ef[i]=0.0;
		for ( j=0 ; j<nbinsout ; ++j ) ef[i]+=square(scalefactor*ma[i][j])*en[j];
	    }
	}
	dAlembertl1=0.0;
	dAlembertl2=0.0;
	dAlembertli=0.0;
	statLi=0.0;
	for ( i=0 ; i<nbinsout ; ++i )
	{
	    incr=fabs(f0[i]-scalefactor*fr[i]);
	    dAlembertl1+=incr;
	    dAlembertl2+=incr*incr;
	    if ( dAlembertli<incr ) dAlembertli=incr;
	    fn[i]+=f0[i]-scalefactor*fr[i];
	    if ( inputErr==NULL && inputerr!=NULL )
	    {
		en[i]+=e0[i]+ef[i];
		if ( statLi<en[i] ) statLi=en[i];
	    }
	}
	dAlembertl2=sqrt(fabs(dAlembertl2));
	/* Do it also for statistical errors. */
	if ( inputErr!=NULL )
	{
	    if ( applytomatrix(nbinsout, nbinsin, (const double**)Errn, compose, &paramcompose, nbinsout, nbinsin, Errr) )
	    {
		dfprintf("[unfold_iterative] Failed in iteration step %d for statistical error calculation.\n", n+1);
		errorcode=-8;
		break;
	    }
	    for ( i=0 ; i<nbinsout ; ++i )
	    {
		unfoldederri=0.0;
		for ( j=0 ; j<nbinsin ; ++j )
		{
		    Errn[i][j]+=Err0[i][j]-scalefactor*Errr[i][j];
		    unfoldederri+=square(Errn[i][j]);
		}
		if ( statLi<unfoldederri ) statLi=unfoldederri;
	    }
	}
	statLi=sqrt(fabs(statLi))/inputpdfL1;
	/* Do it for approximation errors. */
	if ( calcappr )
	{
	    if ( applytomatrix(nbinsout, nbinsout, (const double**)xin, compose, &paramcompose, nbinsout, nbinsout, xir) )
	    {
		dfprintf("[unfold_iterative] Failed in iteration step %d for approximation error calculation.\n", n+1);
		errorcode=-9;
		break;
	    }
	    apprLi=0.0;
	    for ( i=0 ; i<nbinsout ; ++i )
	    {
		xidiffdotunfolded=0.0;
		for ( j=0 ; j<nbinsout ; ++j )
		{
		    xin[j][i]+=xi0[j][i]-scalefactor*xir[j][i];
		    xidiffdotunfolded+=(xi[j][i]-xin[j][i])*fn[j];
		}
		a[i]=fabs(xidiffdotunfolded);
		if ( apprLi<a[i] ) apprLi=a[i];
	    }
	    apprLi/=inputpdfL1;
	}
	/* Do it for systematic errors. */
	if ( inputsys!=NULL || foldopsysT!=NULL )
	{
	    if ( applytomatrix(nbinsout, nbinsout, (const double**)Xin, compose, &paramcompose, nbinsout, nbinsout, Xir) )
	    {
		dfprintf("[unfold_iterative] Failed in iteration step %d for systematic error calculation 1.\n", n+1);
		errorcode=-10;
		break;
	    }
	    for ( i=0 ; i<nbinsout ; ++i )
	    {
		for ( j=0 ; j<nbinsout ; ++j ) Xin[i][j]+=Xi0[i][j]-scalefactor*Xir[i][j];
	    }
	    if ( applytomatrix(nbinsout, nbinsout, (const double**)Xin, compose, &paramcomposeinitialtranspose, nbinsin, nbinsout, Ximod) )
	    {
		dfprintf("[unfold_iterative] Failed in iteration step %d for systematic error calculation 2.\n", n+1);
		errorcode=-11;
		break;
	    }
	    if ( inputsys!=NULL )
	    {
		for ( i=0 ; i<nbinsout ; ++i )
		{
		    Ximoddotinputsys=0.0;
		    for ( j=0 ; j<nbinsin ; ++j )
		    {
			Ximoddotinputsys+=fabs(Ximod[j][i])*fabs(inputsys[j]);
		    }
		    s[i]=scalefactor*Ximoddotinputsys;
		}
	    }
	    else if ( foldopsysT!=NULL )
	    {
		for ( i=0 ; i<nbinsout ; ++i )
		{
		    for ( j=0 ; j<nbinsin ; ++j )
		    {
			Ximod[j][i]=fabs(Ximod[j][i]);
		    }
		}
		if ( applytomatrix(nbinsin, nbinsout, (const double**)Ximod, foldopsysT, &paramssysT, nbinsout, nbinsout, Ximod2) )
		{
		    dfprintf("[unfold_iterative] Failed in iteration step %d for systematic error calculation 3.\n", n+1);
		    errorcode=-12;
		    break;
		}
		for ( i=0 ; i<nbinsout ; ++i )
		{
		    Ximod2Li=0.0;
		    for ( j=0 ; j<nbinsout ; ++j )
		    {
			if ( Ximod2Li<fabs(Ximod2[j][i]) ) Ximod2Li=fabs(Ximod2[j][i]);
		    }
		    s[i]=scalefactor*Ximod2Li*inputpdfL1;
		}
	    }
	    systLi=0.0;
	    for ( i=0 ; i<nbinsout ; ++i )
	    {
		if ( systLi<s[i] ) systLi=s[i];
	    }
	    systLi/=inputpdfL1;
	}
	/* Report number of iterations, maximum aproximation, statistical and systematic errors, and L^1,L^2,L^\infty d'Alembert convergence factors. */
	if ( niterations!=NULL ) *niterations=n+1;
	if ( maxappr!=NULL ) maxappr[n]=apprLi;
	if ( maxstat!=NULL ) maxstat[n]=statLi;
	if ( maxsyst!=NULL ) maxsyst[n]=systLi;
	if ( dAlembertL1!=NULL ) dAlembertL1[n]=(n>0 ? dAlembertl1/dAlembertl1prev : 1.0);
	if ( dAlembertL2!=NULL ) dAlembertL2[n]=(n>0 ? dAlembertl2/dAlembertl2prev : 1.0);
	if ( dAlembertLi!=NULL ) dAlembertLi[n]=(n>0 ? dAlembertli/dAlembertliprev : 1.0);
	dAlembertl1prev=dAlembertl1;
	dAlembertl2prev=dAlembertl2;
	dAlembertliprev=dAlembertli;
	/* Break loop if penalty value becomes too large. */
	if ( verbositylevel>0 )
	{
	    fprintf(stdout, "INFO[unfold_iterative]: Iteration step %d done.\n", n+1); fflush(stdout);
	}
	if ( n>miniteration && stopping!=NULL && ( inputErr!=NULL || inputerr!=NULL || inputsys!=NULL || calcappr ) )
	{
	    if ( stopping(apprLi, statLi, systLi, stopparams) ) break;
	}
    }

    /* Report final result. */
    /* The pdf. */
    if ( unfoldedpdf!=NULL ) copyvector(nbinsout, fn, unfoldedpdf);
    /* The statistical error. */
    if ( inputErr!=NULL && unfoldedErr!=NULL ) copymatrix(nbinsout, nbinsin, (const double**)Errn, unfoldedErr);
    else if ( inputErr!=NULL && unfoldederr!=NULL )
    {
	for ( i=0 ; i<nbinsout ; ++i )
	{
	    unfoldederr[i]=0.0;
	    for ( j=0 ; j<nbinsin ; ++j )
	    {
		unfoldederr[i]+=square(Errn[i][j]);
	    }
	    unfoldederr[i]=sqrt(fabs(unfoldederr[i]));
	}
    }
    else if ( inputerr!=NULL && unfoldederr!=NULL )
    {
	for ( i=0 ; i<nbinsout ; ++i ) unfoldederr[i]=sqrt(fabs(en[i]));
    }
    /* The approximation error. */
    if ( calcappr && unfoldedappr!=NULL ) copyvector(nbinsout, a, unfoldedappr);
    /* The systematic error. */
    if ( ( inputsys!=NULL || foldopsysT!=NULL ) && unfoldedsys!=NULL ) copyvector(nbinsout, s, unfoldedsys);

    /* Free the buffers. */
    if ( f0!=NULL ) { freevector(f0); f0=NULL; }
    if ( fn!=NULL ) { freevector(fn); fn=NULL; }
    if ( fr!=NULL ) { freevector(fr); fr=NULL; }
    if ( Err0!=NULL ) { freematrix(Err0, nbinsout); Err0=NULL; }
    if ( Errn!=NULL ) { freematrix(Errn, nbinsout); Errn=NULL; }
    if ( Errr!=NULL ) { freematrix(Errr, nbinsout); Errr=NULL; }
    if ( e0!=NULL ) { freevector(e0); e0=NULL; }
    if ( en!=NULL ) { freevector(en); en=NULL; }
    if ( ef!=NULL ) { freevector(ef); ef=NULL; }
    if ( ma!=NULL ) { freematrix(ma, nbinsout); ma=NULL; }
    if ( xi!=NULL ) { freematrix(xi, nbinsout); xi=NULL; }
    if ( xi0!=NULL ) { freematrix(xi0, nbinsout); xi0=NULL; }
    if ( xin!=NULL ) { freematrix(xin, nbinsout); xin=NULL; }
    if ( xir!=NULL ) { freematrix(xir, nbinsout); xir=NULL; }
    if ( a!=NULL ) { freevector(a); a=NULL; }
    if ( Xi0!=NULL ) { freematrix(Xi0, nbinsout); Xi0=NULL; }
    if ( Xin!=NULL ) { freematrix(Xin, nbinsout); Xin=NULL; }
    if ( Xir!=NULL ) { freematrix(Xir, nbinsout); Xir=NULL; }
    if ( Ximod!=NULL ) { freematrix(Ximod, nbinsin); Ximod=NULL; }
    if ( Ximod2!=NULL ) { freematrix(Ximod2, nbinsout); Ximod2=NULL; }
    if ( s!=NULL ) { freevector(s); s=NULL; }

    return errorcode;
}



