/**
 * unfold_iterative.h   This is a demo library for unfolding using 
 *                      iterative method of 
 *                      A.Laszlo: J.Phys A39 (2006) 13621
 *                      A.Laszlo: JPCS 368 (2012) 012043
 *                      http://arxiv.org/abs/1404.2787
 *
 *                      Usage and declarations in C follow.
 *
 * Contains usage and declarations for the C implementation:
 * version                  --  Returns version string of the library.
 * verbosity                --  Sets verbosity level of the library.
 * folding                  --  Type definition for folding operators.
 * stoppingfunc             --  Type definition for iteration stopping function.
 * stopatgoodapprox         --  An iteration stopping function.
 * stopbystat               --  An iteration stopping function.
 * stopbysyst               --  An iteration stopping function.
 * stopbystatandsyst        --  An iteration stopping function.
 * stopbycombinederr        --  An iteration stopping function.
 * unfold_scalefactor       --  Procedure to calculate optimal scale factor.
 * unfold_iterative         --  The iterative unfolding procedure.
 *
 */


#ifndef __UNFOLD_ITERATIVE
#define __UNFOLD_ITERATIVE


/** Provide 'extern "C"' instead of 'extern' in case of C++ compilers. */
#ifdef __cplusplus
#define EXTERN extern "C"
#else
#define EXTERN extern
#endif


/**
 * Returns the version string of the library.
 * 
 * @return  version	Pointer to the version string.
 */
EXTERN const char* version(void);


/**
 * Set verbosity level.
 * 
 * @param  level	Integer specifying verbosity (default: 1).
 */
EXTERN void verbosity(const int level);


/**
 * Type for folding operator procedure in any dimensional probability space.
 * This type is a function pointer. The parameters and return value of 
 * such a function is the following.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at 
 *         		nbinsin space point locations.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  nbinsout	Total number of bins in the folded histogram.
 * @param  foldedpdf	Buffer to return output probability density function 
 *         		values sampled at nbinsout locations.
 * @return errorcode	Nonzero on error.
 */
typedef int (*folding)(const int nbinsin, const double* inputpdf, 
		       const void* parameters, 
		       const int nbinsout, double* foldedpdf);


/**
 * Type for iteration stopping function. It is takes the current value of 
 * maximal approximation error, maximal statistical error, maximal systematic 
 * error of reconstructed bin probabilities, and returns nonzero if iteration 
 * stopping criterion is reached.
 * 
 * @param  maxapprox	Maximum approximation error of bin probabilities.
 * @param  maxstat	Maximum statistical error of bin probabilities.
 * @param  maxsyst	Maximum systematic error of bin probabilities.
 * @param  stopparams	Pointer to user defined parameters of stopping function.
 * @return stop		Nonzero on error.
 */
typedef int (*stoppingfunc)(const double maxapprox,
			    const double maxstat,
			    const double maxsyst,
			    const void* stopparams);


/**
 * The simplest iteration stopping function: stop if approximation error becomes 
 * smaller than a threshold.
 */
EXTERN int stopatgoodapprox(const double maxapprox,
			    const double maxstat,
			    const double maxsyst,
			    const void* threshold);


/**
 * Other simple iteration stopping function: stop when propagated 
 * statistical error exceeds approximation error.
 */
EXTERN int stopbystat(const double maxapprox,
		      const double maxstat,
		      const double maxsyst,
		      const void* null);


/**
 * Other simple iteration stopping function: stop when propagated 
 * systematic error exceeds a threshold.
 */
EXTERN int stopbysyst(const double maxapprox,
		      const double maxstat,
		      const double maxsyst,
		      const void* threshold);


/**
 * Other simple iteration stopping function: stop when combined 
 * propagated statistical and systematic error exceeds a threshold.
 */
EXTERN int stopbystatandsyst(const double maxapprox,
			     const double maxstat,
			     const double maxsyst,
			     const void* threshold);


/**
 * Other simple iteration stopping function: stop when combined 
 * propagated statistical, systematic and approximaton error exceeds a threshold.
 */
EXTERN int stopbycombinederr(const double maxapprox,
			     const double maxstat,
			     const double maxsyst,
			     const void* threshold);


/**
 * Procedureto calculate a specific operator norm of the folding operator 
 * to judge convergence in any dimensional probability space. If we denote 
 * the probability base space by X, and the response function by rho, then 
 * the operator norm in question is the
 * \f$
 * sup_{z \in X} \int_{x \in X} \int_{y \in X} 
 *                   rho(y|z) (kappa*rho)(y|x) dy dx
 * \f$
 * expression, where kappa is a smoothing pdf of the form 
 * kappa=P(eta)*eta where eta is a pdf with positive integrable Fourier 
 * transform, P is the parity-reflection operator, and '*' is convolution 
 * (e.g. a Gaussian pdf centered to zero does this job). 
 * Whenever the value of this expression is finite, the iterative 
 * unfolding procedure is convergent.
 * The parameters and return value of this function is the following.
 * 
 * @param  nbinsin	Total number of bins in the input histogram which 
 *         		you want to unfold.
 * @param  foldingop	The folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  foldingopS	The smoother folding operator. It should be a 
 *         		convolution with a pdf of the form P(eta)*eta, 
 *         		where eta is a pdf with positive and integrable 
 *         		Fourier transform, P is the parity-reflection 
 *         		operator, and '*' is convolution. Such a good 
 *         		smoother folding operator is e.g. the convolution by 
 *         		a Gaussian pdf centered to zero. Set it to NULL 
 *         		if you do not want to apply smoother folding for 
 *         		improved convergence.
 * @param  parametersS	Additional parameters passed to the smoother 
 *         		folding operator. Set it to NULL if no smoother 
 *         		folding operator is applied.
 * @param  foldingopT	The folding operator by the transposed response 
 *         		function, i.e. when the two variables of the 
 *         		response function is swapped. Set it to NULL 
 *         		if you do not want to apply transposed folding 
 *         		for improved convergence.
 * @param  parametersT	Additional parameters passed to the transposed 
 *         		folding operator. Set it to NULL if no transposed 
 *         		folding operator is applied.
 * @param  nbinsout	Total number of bins in the unfolded histogram.
 * @param  operatornorm	Pointer to return the computed operator norm.
 * @return errorcode	Nonzero on error.
 */
EXTERN int unfold_scalefactor(const int nbinsin, 
			      const folding foldingop, const void* parameters, 
			      const folding foldingopS, const void* parametersS, 
			      const folding foldingopT, const void* parametersT, 
			      const int nbinsout, 
			      double* operatornorm);


/**
 * Procedurefor unfolding using iterative method in any dimensional 
 * probability space. The parameters and return value of this function 
 * is the following.
 * 
 * @param  nbinsin	Total number of bins in the input histogram.
 * @param  inputpdf	Input probability density function values at 
 *         		nbins space point locations.
 * @param  inputErr	Statistical error matrix of the values of inputpdf 
 *         		(by error matrix we mean any nbinsin x nbinsin 
 *         		matrix Err, for which Err Err^{T} = Covariance holds). 
 *         		Set it to NULL if you don't want exact statistical error 
 *         		propagation. Exact error propagation can be expensive.
 * @param  inputerr	Statistical errors of the values of inputpdf for approximate 
 *         		error propagation with neglection of cross-covariances 
 *         		at each step. Set it to NULL if you don't 
 *         		want aproximate statistical error propagation.
 * @param  inputsys	Systematic errors of the values of inputpdf.
 *         		Set it to NULL if you don't want exact systematic 
 *         		error propagation.
 * @param  foldingop	The folding operator.
 * @param  parameters	Additional parameters passed to the folding operator.
 * @param  foldingopS	The smoother folding operator. It should be a 
 *         		convolution with a pdf of the form P(eta)*eta, 
 *         		where eta is a pdf with positive and integrable 
 *         		Fourier transform, P is the parity-reflection 
 *         		operator, and '*' is convolution. Such a good 
 *         		smoother folding operator is e.g. the convolution by 
 *         		a Gaussian pdf centered to zero. Set it to NULL 
 *         		if you do not want to apply smoother folding for 
 *         		improved convergence.
 * @param  parametersS	Additional parameters passed to the smoother 
 *         		folding operator. Set it to NULL if no smoother 
 *         		folding operator is applied.
 * @param  foldingopT	The folding operator by the transposed response 
 *         		function, i.e. when the two variables of the 
 *         		response function is swapped. Set it to NULL 
 *         		if you do not want to apply transposed folding 
 *         		for improved convergence.
 * @param  parametersT	Additional parameters passed to the transposed 
 *         		folding operator. Set it to NULL if no transposed 
 *         		folding operator is applied.
 * @param  foldopsysT	The transposed systematic errors of folding operator.
 *         		Set it to NULL if you don't want exact systematic 
 *         		error propagation.
 * @param  paramssysT	Additional parameters passed to the transposed 
 *         		folding operator. Set it to NULL if no transposed 
 *         		systematic error folding operator is applied.
 * @param  scalefactor  Scaling factor for improved convergence, which is 
 *         		1 / the value returned by unfold_scalefactor. 
 *         		Set it to 1 if you do not want optimal scaling.
 * @param  miniteration	Minimal number of iterations.
 * @param  maxiteration	Maximal number of iterations.
 * @param  calcappr	Set it to 1 if you want calculation of approximation 
 *         		error.
 * @param  stopping	Stopping function to decide iteration stopping based 
 *         		on approximation, statistical and systematic errors. 
 *         		Set it to NULL if you don't want iteration stopping.
 * @param  stopparams	Parameters for the above stopping function.
 *         		Set it to NULL if stopping not used.
 * @param  nbinsout	Total number of bins in the unfolded histogram.
 * @param  unfoldedpdf	Output probability density function values at 
 *                      nbinsout sampling points.
 * @param  unfoldedErr	Statistical error matrix of the values of unfoldedpdf 
 *         		(by error matrix we mean any nbinsout x nbinsin 
 *         		matrix Err, for which Err Err^{T} = Covariance holds). 
 *         		Set it to NULL if you don't want exact statistical error 
 *         		propagation.
 * @param  unfoldederr	Statistical errors of the values of unfoldedpdf after approximate 
 *         		error propagation with neglection of cross-covariances. 
 *         		Set it to NULL if you don't want approximate statistical error 
 *         		propagation.
 * @param  unfoldedappr	Approximation errors of the values of the unfoldedpdf. 
 *         		Set it to NULL if you don't want approximation errors.
 * @param  unfoldedsys	Systematic errors of the values of the unfoldedpdf. 
 *         		Set it to NULL if you don't want systematic errors.
 * @param  niterations	Buffer to return performed number of iterations. 
 *         		Set it to NULL if not interested in.
 * @param  maxappr	Buffer to return values of maximum approximation error 
 *         		for bin probability values into a buffer pre-allocated 
 *         		to be able to contain maxiteration number of values. 
 *         		Set it to NULL if not interested in.
 * @param  maxstat	Buffer to return values of maximum statistical error 
 *         		for bin probability values into a buffer pre-allocated 
 *         		to be able to contain maxiteration number of values. 
 *         		Set it to NULL if not interested in.
 * @param  maxsyst	Buffer to return values of maximum systematic error 
 *         		for bin probability values into a buffer pre-allocated 
 *         		to be able to contain maxiteration number of values. 
 *         		Set it to NULL if not interested in.
 * @param  dAlembertL1  Buffer to return L^1 d'Alembert convergence factors 
 *         		into a buffer pre-allocated to be able to contain 
 *         		maxiteration number of values. Set it to NULL if 
 *         		you are not interested in the L^1 d'Alembert factors.
 * @param  dAlembertL2  Buffer to return L^2 d'Alembert convergence factors 
 *         		into a buffer pre-allocated to be able to contain 
 *         		maxiteration number of values. Set it to NULL if 
 *         		you are not interested in the L^2 d'Alembert factors.
 * @param  dAlembertLi  Buffer to return L^infinity d'Alembert convergence 
 *         		factors into a buffer pre-allocated to be able to 
 *         		contain maxiteration number of values. Set it to 
 *         		NULL if you are not interested in the L^infinity 
 *         		d'Alembert factors.
 * @return errorcode	Nonzero on error.
 */
EXTERN int unfold_iterative(const int nbinsin, const double* inputpdf, 
			    const double** inputErr, const double* inputerr, const double* inputsys, 
			    const folding foldingop, const void* parameters, 
			    const folding foldingopS, const void* parametersS, 
			    const folding foldingopT, const void* parametersT, 
			    const folding foldopsysT, const void* paramssysT, 
			    const double scalefactor, 
			    const int miniteration, 
			    const int maxiteration, 
			    const int calcappr, 
			    const stoppingfunc stopping, const void* stopparams, 
			    const int nbinsout, double* unfoldedpdf, 
			    double** unfoldedErr, double* unfoldederr, 
			    double* unfoldedappr, double* unfoldedsys, 
			    int *niterations,
			    double* maxappr, double* maxstat, double* maxsyst, 
			    double* dAlembertL1, double* dAlembertL2, double* dAlembertLi);


#endif /* __UNFOLD_ITERATIVE */
